/*
 * Copyright (c) 2006, Adam Dunkels
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
#ifndef __TOKENIZER_H__
#define __TOKENIZER_H__

typedef enum {
  TOKENIZER_ERROR,
  TOKENIZER_CR,
  TOKENIZER_ENDOFINPUT,
  TOKENIZER_VARIABLE, // 3
  TOKENIZER_PLUS,
  TOKENIZER_MINUS,
  TOKENIZER_AND,
  TOKENIZER_OR,
  TOKENIZER_XOR,
  TOKENIZER_ASTR,
  TOKENIZER_SLASH,
  TOKENIZER_LEFTPAREN,
  TOKENIZER_RIGHTPAREN, // 12
  TOKENIZER_LT,
  TOKENIZER_GT,
  TOKENIZER_EQ,
  TOKENIZER_NE,
  TOKENIZER_LE,
  TOKENIZER_GE,
  TOKENIZER_NOT,
  TOKENIZER_LNOT,
  TOKENIZER_LAND,
  TOKENIZER_LOR,
  TOKENIZER_NOR,
  TOKENIZER_MOD,
  TOKENIZER_COMMA, // 25
  TOKENIZER_SEMICOLON,
  TOKENIZER_SEP,
  TOKENIZER_ODD, // 1,3,5,7,9,...
  TOKENIZER_ABS,
  TOKENIZER_NUMBER, // 30
  TOKENIZER_STRING,
  TOKENIZER_LABEL, // :label | 100*i+x 
  TOKENIZER_EXEC, // switch '{'  mulinecomment to code
  TOKENIZER_DATA, // verry flexibel 
  TOKENIZER_IF,
  TOKENIZER_THEN,
  TOKENIZER_ELSEIF,  // 37
  TOKENIZER_ELSE,
  TOKENIZER_ENDIF,
  TOKENIZER_BREAK,
  TOKENIZER_CONT,
  TOKENIZER_SELECT, // 42
  TOKENIZER_CASE_ELSE,
  TOKENIZER_CASE,
  TOKENIZER_END_SELECT,
  TOKENIZER_FOR,
  TOKENIZER_TO, // if x in 15..42 then
  TOKENIZER_STEP,
  TOKENIZER_NEXT,
  TOKENIZER_GOTO, // 50
  TOKENIZER_GOSUB, // calculated jump method: gosub 100*i+x
  TOKENIZER_RETURN,// see data statement;....4..
  TOKENIZER_SKIP,
  TOKENIZER_DO,
  TOKENIZER_UNTIL,
  TOKENIZER_WHILE,
  TOKENIZER_WEND,
  TOKENIZER_LET,
  TOKENIZER_PRINT_SCREEN, // == Pri.
  TOKENIZER_PRINT, // 60
  TOKENIZER_REM,
  TOKENIZER_CLS,
  TOKENIZER_CLICK,
  TOKENIZER_PRESS,
  TOKENIZER_RELEASE,
  TOKENIZER_SHOOT, // "shoot ~" should be rawproof, Caefix
  TOKENIZER_GET_SHOOTING,
  TOKENIZER_IS_KEY,
  TOKENIZER_IS_PRESSED,
  TOKENIZER_WAIT_CLICK, // 70
  TOKENIZER_SET_AELOCK,
  TOKENIZER_SET_AFLOCK,
  TOKENIZER_MD_DETECT_MOTION, // 73
  TOKENIZER_MD_GET_CELL_DIFF,
  TOKENIZER_MD_GET_CELL_VAL,
  TOKENIZER_MD_DETECT_TICK,
  TOKENIZER_WRITE,
  TOKENIZER_ECHO,
  TOKENIZER_EXIT,  // between goto & reset
  TOKENIZER_SQRT,
  TOKENIZER_IN,   // if x in 1,3[,5,7,x+y...]
  TOKENIZER_EVEN, // 0,2,4,6,8,...82
  
  TOKENIZER_GET_ORIENTATION_SENSOR,
  TOKENIZER_GET_FREE_DISK_SPACE,
  TOKENIZER_GET_TV96,
  TOKENIZER_GET_USER_TV96,
  TOKENIZER_GET_USER_TV_ID,
  TOKENIZER_SET_TV96_DIRECT,
  TOKENIZER_SET_TV96,
  TOKENIZER_SET_USER_TV96,
  TOKENIZER_SET_USER_TV_BY_ID_REL,
  TOKENIZER_SET_USER_TV_BY_ID,
  TOKENIZER_GET_AV96,
  TOKENIZER_GET_CURRENT_AV96,
  TOKENIZER_GET_CURRENT_TV96,
  TOKENIZER_GET_CURRENT_BASE_SV96,
  TOKENIZER_GET_CURRENT_DELTA_SV96,
  TOKENIZER_GET_MAX_AV96,
  TOKENIZER_GET_MIN_AV96,
  TOKENIZER_GET_USER_AV96,
  TOKENIZER_GET_USER_AV_ID,
  TOKENIZER_SET_AV96_DIRECT,
  TOKENIZER_SET_AV96,
  TOKENIZER_SET_USER_AV96,
  TOKENIZER_SET_USER_AV_BY_ID_REL,
  TOKENIZER_SET_USER_AV_BY_ID,
  //TOKENIZER_SET_SHUTTER_SPEED,
  TOKENIZER_GET_DAY_SECONDS,
  TOKENIZER_GET_TICK_COUNT,
  TOKENIZER_GET_TIME, // x=get_time (>6) prints timestamp 
  TOKENIZER_SET_MF,
  TOKENIZER_GET_AFLOCK,
  TOKENIZER_GET_AF_FRAME,
  TOKENIZER_GET_SD_OVER_MODES,
  TOKENIZER_GET_EFF_FOCAL_LENGTH,
  TOKENIZER_GET_FOCAL_LENGTH,
  TOKENIZER_GET_FOCUS_STATE,
  TOKENIZER_GET_FOCUS_MODE,
  TOKENIZER_GET_FOCUS_OK,
  TOKENIZER_GET_FOCUS,
  TOKENIZER_SET_FOCUS,
  TOKENIZER_GET_NEAR_LIMIT,
  TOKENIZER_GET_FAR_LIMIT,
  TOKENIZER_GET_DOF,
  TOKENIZER_GET_HYPERFOCAL_DIST,
  TOKENIZER_GET_MIN_STACK_DIST,
  TOKENIZER_GET_SD_OVER_VALUE,
  TOKENIZER_IS_INFINITY,
  TOKENIZER_OUT_OF_FOCUS,
  TOKENIZER_GET_ZOOM_STEPS,
  TOKENIZER_GET_ZOOM,
  TOKENIZER_SET_ZOOM_SPEED,
  TOKENIZER_SET_ZOOM_REL,
  TOKENIZER_SET_ZOOM,
  TOKENIZER_SET_RECORD,
  TOKENIZER_GET_ISO_MARKET,
  TOKENIZER_GET_ISO_REAL,
  TOKENIZER_GET_ISO_MODE,
  //TOKENIZER_SET_ISO_DL_F_B,
  //TOKENIZER_SET_ISO_MARKET,
  TOKENIZER_SET_ISO_REAL,
  TOKENIZER_SET_ISO_MODE,
  TOKENIZER_SET_ISO_DIRECT,
  TOKENIZER_SET_EV,
  TOKENIZER_GET_EV,
//  TOKENIZER_GET_LUMINANCE,
  TOKENIZER_GET_BV96,
  TOKENIZER_GET_SV96,
  TOKENIZER_SET_SV96,
  TOKENIZER_GET_PROP,
  TOKENIZER_SET_PROP,
  TOKENIZER_SET_LED,
  TOKENIZER_GET_RAW_SUPPORT,
  TOKENIZER_SET_RAW_NR,
  TOKENIZER_GET_RAW_NR,
  TOKENIZER_SET_RAW,
  TOKENIZER_GET_RAW,
  TOKENIZER_GET_JPG_COUNT,
  TOKENIZER_GET_RAW_COUNT,
  TOKENIZER_GET_EXP_COUNT,
  TOKENIZER_GET_FLASH_MODE,
  TOKENIZER_GET_FLASH_READY,
  TOKENIZER_GET_MOVIE_STATUS,
  TOKENIZER_SET_MOVIE_STATUS,
  TOKENIZER_GET_VIDEO_RECORDING,
  TOKENIZER_GET_VIDEO_BUTTON,
  TOKENIZER_GET_RESOLUTION,
  TOKENIZER_SET_RESOLUTION,
  TOKENIZER_GET_QUALITY,
  TOKENIZER_SET_QUALITY,
  TOKENIZER_GET_MODE,
  TOKENIZER_GET_METERING_MODE,
  TOKENIZER_SET_METERING_MODE,
  TOKENIZER_GET_IMAGER_ACTIVE,
  TOKENIZER_GET_CANON_RAW_SUPPORT,
  TOKENIZER_GET_CANON_IMAGE_FORMAT,
  TOKENIZER_SET_CANON_IMAGE_FORMAT,
  TOKENIZER_GET_CAPTURE_MODE,
  TOKENIZER_SET_CAPTURE_MODE_CANON,
  TOKENIZER_SET_CAPTURE_MODE,
  TOKENIZER_SET_AF_FRAME,
  TOKENIZER_IS_CAPTURE_MODE_VALID,
  TOKENIZER_GET_CONFIG_VALUE,
  TOKENIZER_SET_CONFIG_VALUE,
  TOKENIZER_SET_YIELD,
  TOKENIZER_ISO_TO_SV96,
  TOKENIZER_SV96_TO_ISO,
  TOKENIZER_ISO_REAL_TO_MARKET,
  TOKENIZER_ISO_MARKET_TO_REAL,
  TOKENIZER_SV96_REAL_TO_MARKET,
  TOKENIZER_SV96_MARKET_TO_REAL,
  TOKENIZER_APERTURE_TO_AV96,
  TOKENIZER_AV96_TO_APERTURE,
  TOKENIZER_USEC_TO_TV96,
  TOKENIZER_TV96_TO_USEC,
  TOKENIZER_SECONDS_TO_TV96,
  TOKENIZER_MD_AF_LED_CONTROL,
  TOKENIZER_RANDOM,
  TOKENIZER_PLAY_SOUND,
  TOKENIZER_SET_EXIT_KEY,
  TOKENIZER_WHEEL_LEFT,
  TOKENIZER_WHEEL_RIGHT,
  TOKENIZER_GET_WHEEL,
  
  TOKENIZER_EXIT_ALT,
  TOKENIZER_ENTER_ALT, 
  TOKENIZER_GET_ALT_MODE,
  TOKENIZER_MEM_DUMP,
  TOKENIZER_GET_DIGIC,
  TOKENIZER_GET_PLATFORM_ID,
  TOKENIZER_SCRIPT_AUTOSTARTED,
  TOKENIZER_GET_SCRIPT_AUTOSTART,
  TOKENIZER_SET_SCRIPT_AUTOSTART,
  TOKENIZER_GET_IS_MODE,
  TOKENIZER_GET_TEMPERATURE,
  TOKENIZER_GET_ND_CURRENT_EV96,
  TOKENIZER_GET_ND_VALUE_EV96,
  TOKENIZER_GET_ND_PRESENT,
  TOKENIZER_SET_ND_FILTER,
  TOKENIZER_GET_PROPSET,
  TOKENIZER_SHOT_HISTO_ENABLE,
  TOKENIZER_GET_HISTO_RANGE,
  TOKENIZER_GET_VBATT,
  TOKENIZER_GET_USB_POWER,
  TOKENIZER_SET_REMOTE_TIMING,
  TOKENIZER_FORCE_ANALOG_AV,
  TOKENIZER_FORCE_USB_PRESENT,
  TOKENIZER_USB_SYNC_WAIT,
  TOKENIZER_SET_CONSOLE_LAYOUT,
  TOKENIZER_SET_CONSOLE_AUTOREDRAW,
  TOKENIZER_CONSOLE_REDRAW,
  TOKENIZER_GET_DISPLAY_MODE,
  TOKENIZER_SET_BACKLIGHT,
  TOKENIZER_SET_LCDDISPLAY,
  TOKENIZER_SET_DRAW_TITLE_LINE,
  TOKENIZER_GET_DRAW_TITLE_LINE,
  TOKENIZER_GET_DRIVE_MODE, 
  TOKENIZER_GET_DISK_SIZE,
  TOKENIZER_SWAP_PARTITIONS,
  TOKENIZER_SHUT_DOWN,
  TOKENIZER_REBOOT,
  TOKENIZER_SLEEP,
  // most () ignored now for more  compat. to some Lualines. 
  // Caefix, 2019
  TOKENIZER_BEGIN, // new first_ptr, upper :labels are lost.
  TOKENIZER_ON, // =1 | on X=0,1,2,3,X goto/gosub 0,10,20,333,"default"
  TOKENIZER_OFF,// =0
  TOKENIZER_TRUE,
  TOKENIZER_FALSE,
  TOKENIZER_IFKEY,
  TOKENIZER_SGN,
  TOKENIZER_UNS,  // (unsigned short) x
  TOKENIZER_CAP,  // while (abs(r)>9) {r=r/10;}; '\-4210'=-4 
  TOKENIZER_LXOR, // a+b=1
  TOKENIZER_NAND, // a && !b
  TOKENIZER_SHL,  // x=y<<z
  TOKENIZER_SHR,
  TOKENIZER_PEEK,
  TOKENIZER_POKE,
  TOKENIZER_VARLIST,
  TOKENIZER_LIST, 
  TOKENIZER_CLEAN, // set vars to 0
  TOKENIZER_SWAP, // i,j ; k,l ; #7,m...
  TOKENIZER_SHIFT, // 3[,7] << data Abakus
  TOKENIZER_MSEC,  // since start
  TOKENIZER_LSEC, // linetime, looptime 
  TOKENIZER_BEEP,   //  ''
  TOKENIZER_MOMENT, //  '.'
  TOKENIZER_SEKUNDE,//  "'"
  TOKENIZER_MINUTE,
  TOKENIZER_MIN,
  TOKENIZER_MAX,
  TOKENIZER_STUNDE,
  TOKENIZER_RESET, // 
  TOKENIZER_LOOP, // 1 start again if error or eof // if|while loop...
  TOKENIZER_TEST,
  TOKENIZER_TRONLY,
  TOKENIZER_TRON,
  TOKENIZER_TROFF,
  TOKENIZER_DEBUG, //  
  TOKENIZER_UBASIC_READER, // display wie goto  
  TOKENIZER_STRLEN,
  TOKENIZER_EXCEPTION, // Romlog ~20..60 secs later
  TOKENIZER_BUG, // ?bug,bug returns  linenr
  TOKENIZER_END // 278 
} ubasic_token; 

void  tokenizer_init(const char *program);
void  tokenizer_next(void);
short tokenizer_token(void);
long  tokenizer_value(void);
short tokenizer_variable_num(void);
void  tokenizer_string(char *dest, int len);
void  tokenizer_label(char *dest, int len);

unsigned short tokenizer_line_number(void);

short tokenizer_finished(void);
void  tokenizer_error_print(void);

/// Caefix 2019-2021
short is_cr();
short is_eof();
short is_comma();
short is_token(short token);
short tokenizer_char(void);
void  tokenizer_begin(void);
void  tokenizer_eol(void);
void  tokenizer_inc(void);
void  tokenizer_sep(void);
void  tokenizer_skip(short token);
void  tokenizer_exec(void); // must have a '{' same line
void  tokenizer_statement(char *dest, int len);
unsigned short tokenizer_reader_pos(int lines);
extern char  subtitle[45]; // gcc10+ needs extern
void  tokenizer_end(void);
#endif /* __TOKENIZER_H__ */

/*
--- < ~~~~~ >
--- <  \./~ >
--- <   I   >
--[===========[
@title !uBaswap.lua
@chdk_version 1.4
@subtitle :...,....1....,....2....,....3....,....4..
--]===========]


os.rename("A/CHDK/MODULES/!uBasic.flt","A/CHDK/MODULES/!!uBasiX.flt")
os.rename("A/CHDK/MODULES/uBasic.flt","A/CHDK/MODULES/!uBasic.flt")
os.rename("A/CHDK/MODULES/!!uBasiX.flt","A/CHDK/MODULES/uBasic.flt")

*/