
/*
 * Copyright (c) 2006, Adam Dunkels
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#if DEBUG
#define DEBUG_PRINTF(...)  printf(__VA_ARGS__)
#else
#define DEBUG_PRINTF(...)
#endif

#include "camera_info.h"
#include "shot_histogram.h"
#include "levent.h"
#include "script.h"
#include "console.h"
#include "conf.h"
#include "lang.h"
#include "gui_lang.h"
#include "action_stack.h"
#include "shooting.h"
#include "sd_card.h"
#include "backlight.h"
#include "battery.h"
#include "temperature.h"
#include "clock.h"
#include "file_counter.h"
#include "lens.h"
#include "debug_led.h"
#include "keyboard.h"
#include "usb_remote.h"
#include "shutdown.h"
#include "sound.h"
#include "motion_detector.h"

#include "string.h"
#include "time.h"
#include "math.h"

#ifdef UBASIC_TEST
#include "camera_functions.h"
Conf conf;                      // TODO: initialisation
_cam_info camera_info;          // TOTO: initialisation
#else
#include "script_key_funcs.h"
#include "properties.h"
#endif

#include "ubasic.h"
#include "tokenizer.h"


// Forward references
//static long ubasic_get_variable(short varnum);
static short ubasic_get_key_arg();
// Variable number to store return values from Action Stack functions (e.g. motion_detect, shoot)
static unsigned short ubasic_as_ret_var_num;

#define INCLUDE_OLD_GET__SYNTAX

static char const *program_ptr;
#define MAX_STRINGLEN 128// Caefix, :...,....1....,....2....,....3....,....4..
static char string[MAX_STRINGLEN];
static char buf[MAX_STRINGLEN];
static char alog[36]; // write appends to "A/CHDK/LOGS/Log_%04d.Log" || write [=["A/Romlog.log"],]....
static char scripttitle[45]=":)";
       char subtitle[45];

#define MAX_IF_STACK_DEPTH 6
static short if_stack[MAX_IF_STACK_DEPTH];
static int if_stack_ptr;

struct select_state {
  int select_value;
  short case_run;
};
#define MAX_SELECT_STACK_DEPTH 5
static struct select_state select_stack[MAX_SELECT_STACK_DEPTH];
static short select_stack_ptr;

#define MAX_GOSUB_STACK_DEPTH 10
static unsigned short gosub_stack[MAX_GOSUB_STACK_DEPTH];
static short gosub_stack_ptr;

#define MAX_BREAK_STACK_DEPTH 36 // Caefix, :...,....1....,....2....,....3....,....4..
static unsigned short break_stack[MAX_BREAK_STACK_DEPTH];
static short break_stack_ptr;

#define MAX_DO_STACK_DEPTH 6
static unsigned short do_stack[MAX_DO_STACK_DEPTH];
static short do_stack_ptr;

#define MAX_WHILE_STACK_DEPTH 5
static unsigned short while_stack[MAX_WHILE_STACK_DEPTH];
static short while_stack_ptr;

struct for_state {
  unsigned short line_after_for;
  unsigned short for_variable;
  int to;
  int step;
};
#define MAX_FOR_STACK_DEPTH 5
static struct for_state for_stack[MAX_FOR_STACK_DEPTH];
static short for_stack_ptr;

#define MAX_VARNUM 64 // ++ #0 - #9 + ## + #@=loop 
static long variables[MAX_VARNUM];


static short ended, end_wait, loop, trace, traconly, debug, ifkey=0, on_go, go_fwd, jogdial, logfile=0;
static int lsec, sec=0;
static short str_len, pause, mem_key=0;
static int stm_ctr=0, subj_dist=0;
static unsigned start_sd;
static unsigned first_tick, start_tick;
static char const on=1, off=0; // ++ on | off 
/*---- 144 ------------------------------------------------------------------*/

// Forward references
static long varfactor(unsigned short n);
static long expr(void);
static long relation(void);
static unsigned get_time(int tmode);
static unsigned wait_click_statement(void);
static void line_statement(void);
static void statement(void);
static void do_statement(void);
static void data_statement(void);
static void print_statement(short token);
static void break_statement(void);
static void moment_statement(unsigned d);
static short  set_mf_statement(void);
//static unsigned stopwatch(int v); 

/*
ubasic scheduling controls
ubasic returns control to kbd_task as soon as either of the following is is exceeded
defaults is 1 ubasic line per kbd_task cycle, mimicing old behavior
*/
#define YIELD_MAX_LINES_DEFAULT 1
#define YIELD_MAX_MS_DEFAULT 10
// maximum number lines before yielding (labels, blank lines and REMs excluded)
static unsigned short yield_max_lines;
// maximum number of ticks before yielding
static unsigned short yield_max_ms;
// set flag to immediately transfer execution flow to main CHDK to process external action
// TODO a better approach to this might be to check if AS_RUN is on top of the action_stack
static short flag_yield;

short ubasic_error;
const char *ubasic_errstrings[UBASIC_E_ENDMARK] =
{
    "No err",
    "Parse err",
    "Unk stmt",
    "Unk key",
    "Unk label",
    "gosub: Stack ful",
    "bad return",
    "if: Stack ful",
    "bad endif",
    "select: Stack ful",
    "bad end_select",
    "for: Stack ful",
    "bad next",
    "do: Stack ful",
    "bad until",
    "while: Stack ful",
    "bad wend",
    "Unk err" 
};

/*---------------------------------------------------------------------------* /
void extern_void(void (*func)(void))
{
    func();
}
/*----- 205 -----------------------------------------------------------------*/
static void reset(void)
{
  ended = 0;
  trace= off; // [tron | troff ]
  traconly=off; // tronly
  jogdial=0;
  flag_yield = 0;
  for_stack_ptr = gosub_stack_ptr = while_stack_ptr = do_stack_ptr = if_stack_ptr = select_stack_ptr = 0;
  break_stack_ptr=0;
  ubasic_error = UBASIC_E_NONE;
  tokenizer_init(program_ptr);
  if (!stm_ctr) {
     tokenizer_statement(scripttitle,45);}
  variables[MAX_VARNUM-1]=loop;
  string[MAX_STRINGLEN-1]='\0';
  buf[MAX_STRINGLEN-1]='\0';
  string[0]='\0';
  alog[35]='\0';
  buf[0]='\0';
  str_len=0;
  go_fwd=0;
  on_go=0;
  pause=0;
  lsec=0;
}

/*----- 233 -----------------------------------------------------------------*/
int ubasic_init(const char *program, __attribute__ ((unused))int is_ptp)
{
  program_ptr = program;
  kbd_get_autoclicked_key();
  first_tick = get_tick_count();
  start_sd = GetFreeCardSpaceKb();
  yield_max_ms = YIELD_MAX_MS_DEFAULT;
  yield_max_lines = YIELD_MAX_LINES_DEFAULT;
  tokenizer_end(); // clean some stuff evtl. left from last runned script. Without Vars A..#9 !!
  loop = off; // reset if error or eof : "loop 1|x" // if loop ...
  debug= off;
  end_wait=0;
  stm_ctr=0;
  reset();
  play_sound(5);
  moment_statement(11);
  return 1;
}
/*---------------------------------------------------------------------------*/
unsigned short ubasic_linenumber();
inline unsigned short ubasic_linenumber()
{  return tokenizer_line_number();}

/*---------------------------------------------------------------------------*/
void reader_init()  // TOKENIZER_UBASIC_READER
{	// limits: linelen 127, 65432 lines, memory
	ended=0; pause=-1; traconly=100;
    play_sound(0);
    console_set_layout(0,0,45,12);
    tokenizer_reader_pos(-1);
    if (stm_ctr==0)  // not ubasic script
	   {go_fwd=-2; pause=5;} 
}
/*---------------------------------------------------------------------------*/
static void accept(short token)
{
  if (ended) return;
  if (token) {//  0 = Joker
  short t=tokenizer_token();
  if (token !=t) {
     if (token<0) {  if (token == -t) {tokenizer_next();}
         return;   } // == if (is_token(-token)) tokenizer_next();
    DEBUG_PRINTF("Token not what was expected (Expected %d, got %d)\n", token, t);
    tokenizer_error_print();
	sprintf(buf, "%d-> Expected %d, got %d",ubasic_linenumber(),token,t);
	script_console_add_error((long)buf); buf[0]='\0';
	moment_statement(99);
	if (!trace) {
		ended = 1;
		ubasic_error = UBASIC_E_PARSE;}
//	if (t==TOKENIZER_CR) tokenizer_sep();
    return;  }
  DEBUG_PRINTF("Expected %d, got it\n", t);
  }//  0 
  tokenizer_next();
  if (is_eof()) {if (!trace) {ended = 1; traconly=off;} return;}
}
/*---------------------------------------------------------------------------*/
static void accept_cr()
{	if (ended) return;
	short e=0;
	short t=tokenizer_token();
  while (t != TOKENIZER_CR) {
	switch(t) { // 
	  case TOKENIZER_SEP:   if (!e) return;  // Caefix, 6.4.19)
	  case TOKENIZER_COMMA:
		   tokenizer_next(); 
		   while (is_cr()) {accept(0);} break;
	  case TOKENIZER_ELSE:
	  case TOKENIZER_ELSEIF:  e=9;  break;
	  case TOKENIZER_END: tokenizer_eol(); // Compat Lua.
           if (!e) {tokenizer_sep();} break;
	  case TOKENIZER_DO:
		 {tokenizer_eol(); do_statement(); break_statement(); tokenizer_sep();}
	  case TOKENIZER_ENDOFINPUT:
		 {ended = 1; traconly=off;  return;}
	  case TOKENIZER_LET:	if (!e) return;  // experimental let
	} // 
	tokenizer_next(); 
	t=tokenizer_token();
  } // while
}
/*---------------------------------------------------------------------------*/
static char eos()
{
	short t = tokenizer_token();
	if (t == TOKENIZER_CR
	 || t == TOKENIZER_SEP
	 || t == TOKENIZER_THEN
	 || t == TOKENIZER_ELSE
	 || t == TOKENIZER_ELSEIF
	 || t == TOKENIZER_END) //  (compat Lua: if o==0 then O=0 end) 
	 { return 1; }
	return 0;
}
/*---------------------------------------------------------------------------* /
static void accept_cr()
{
    while(tokenizer_token() != TOKENIZER_CR &&
	    tokenizer_token() != TOKENIZER_ENDOFINPUT) tokenizer_next();
    accept(TOKENIZER_CR);
}
/ *---------------------------------------------------------------------------*/
static int shooting_get_near_limit_of_acceptable_sharpness()
{
  shooting_update_dof_values();
  return camera_info.dof_values.near_limit;
}
static int shooting_get_far_limit_of_acceptable_sharpness()
{
  shooting_update_dof_values();
  return camera_info.dof_values.far_limit;
}
static int shooting_get_depth_of_field()
{
  shooting_update_dof_values();
  return camera_info.dof_values.depth_of_field;
}
static int shooting_get_min_stack_distance()
{
  shooting_update_dof_values();
  return camera_info.dof_values.min_stack_distance;
}
/*---------------------------------------------------------------------------*/
static short varlist(void)
{
short i,n=0; long val;
	sprintf(buf, "%s",scripttitle);
	script_console_add_error((long)buf);
strcpy(string,"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789#@");
///            a:...,....1....,....2....,....3....,....4....,....5....,....6..@ 
for (i=0; i<MAX_VARNUM; ++i) {
    buf[0]=0;
    val = variables[i];
	if (val!=0) { n++;
		sprintf(buf, "#%c= %d",*(string+i),val);
		script_console_add_error((long)buf);}
	}   
    //i = conf.platformid;
	sprintf(buf, "@line:%d :%d:: %d",tokenizer_line_number(), gosub_stack_ptr, n);//,"  PID_",i);
    script_console_add_error((long)buf);
	moment_statement(96);
	get_time(7);
	get_time(8);
	buf[0]='\0';
return n; // count vars !=0
}
/*---------------------------------------------------------------------------*/
void        ubasic_set_variable(unsigned short varnum, long value);
inline void ubasic_set_variable(unsigned short varnum, long value) {
//  if(varnum >= 0)// && varnum < MAX_VARNUM) 
    variables[varnum&63] = value; //else ended=1;
}


/*---------------------------------------------------------------------------*/
static int factor(void)
{
  int r = 0;
  tConfigVal configVal;

  register short t = tokenizer_token();
  if (t == TOKENIZER_STRING) { 
      r=ubasic_get_key_arg();
	  if (!ubasic_error) return r;
	  ubasic_error=UBASIC_E_NONE; // ignore
  t=tokenizer_token();}
  DEBUG_PRINTF("factor: token %d\n", tokenizer_token());
  switch(t) {
  case TOKENIZER_VARIABLE:
    r = varfactor(MAX_VARNUM);
    break;
  case TOKENIZER_NUMBER:
    r = tokenizer_value();
    DEBUG_PRINTF("factor: number %d\n", r);
    accept(TOKENIZER_NUMBER);
    break;
  case TOKENIZER_MOD: //  %0..%9 ,Args like in Batchfiles; for data and shift
    r = tokenizer_variable_num();
    if (r<0) {accept(TOKENIZER_VARIABLE); break;};
    r = variables[r];
    tokenizer_next();
    break;
  case TOKENIZER_PLUS:
    accept(TOKENIZER_PLUS);  r = factor();
    break;
  case TOKENIZER_MINUS:
    accept(TOKENIZER_MINUS); r = -factor();
    break;
  case TOKENIZER_NOT:
    accept(TOKENIZER_NOT);   r = ! factor();
    break;
  case TOKENIZER_LNOT:
    accept(TOKENIZER_LNOT);  r = ! relation();
    break;
  case TOKENIZER_LEFTPAREN:
    accept(TOKENIZER_LEFTPAREN);
    r = relation();
    accept(TOKENIZER_RIGHTPAREN);
    break;

  case TOKENIZER_ABS:
    accept(TOKENIZER_ABS);
    r = factor(); if (r<0) {r=-r;};
    break;
  case TOKENIZER_SGN:
    accept(TOKENIZER_SGN);
    r = factor();
    if (r<0) {r=-1;}; if (r>0) {r=1;};
    break;
  case TOKENIZER_UNS:
    accept(TOKENIZER_UNS); 
    r = (unsigned short)factor();
    break;
  case TOKENIZER_ODD:
    accept(TOKENIZER_ODD); 
    r = factor()&1;
    break;
  case TOKENIZER_EVEN:
    accept(TOKENIZER_EVEN); 
    r = 1-(factor()&1);
    break;
  case TOKENIZER_CAP:
    accept(TOKENIZER_CAP);
    r = factor(); while (abs(r)>9) {r=r/10;};
    break;
  case TOKENIZER_MAX: r=-0xfffffff;
    do {accept(0);
       if (is_cr()) accept(0);
       int v=expr(); if (v>r) r=v;} // 
    while (is_comma());
    accept(-TOKENIZER_SEMICOLON);
    break;
  case TOKENIZER_MIN: r=0xfffffff;
    do {accept(0);
       if (is_cr()) accept(0);
       int v=expr(); if (v<r) r=v;} // 
    while (is_comma());
    accept(-TOKENIZER_SEMICOLON);
    break;
  case TOKENIZER_SQRT:
    accept(TOKENIZER_SQRT);
    r = (long)sqrt((double)factor()); 
    break;

  case TOKENIZER_OFF:
    accept(TOKENIZER_OFF); break;
  case TOKENIZER_ON:
    accept(TOKENIZER_ON);
      r=1;	break;

  case TOKENIZER_TRON:
    accept(TOKENIZER_TRON);
      r=trace;	break;
  case TOKENIZER_TRUE:
    accept(TOKENIZER_TRUE);
      r=1;	break;
  case TOKENIZER_FALSE:
    accept(TOKENIZER_FALSE); break;
	
  case TOKENIZER_GET_SHOOTING:
    accept(TOKENIZER_GET_SHOOTING);
    r = shooting_in_progress();
    break;
  case TOKENIZER_GET_TICK_COUNT:
    accept(TOKENIZER_GET_TICK_COUNT);
    r = get_tick_count();     
	flag_yield=1;
    break;
  case TOKENIZER_IS_KEY:
    accept(TOKENIZER_IS_KEY);
//	moment_statement(20);  // Turns Cam OFF // Why?
	if (ifkey<0) r = script_key_is_pressed(ubasic_get_key_arg());
    else r = script_key_is_clicked(ubasic_get_key_arg());
//	moment_statement(20); ... OFF // no need, but strange.
	flag_yield=1;
    break;
  case TOKENIZER_IS_PRESSED:
    accept(TOKENIZER_IS_PRESSED);
    if (ifkey<1) r = script_key_is_pressed(ubasic_get_key_arg());
    else r = script_key_is_clicked(ubasic_get_key_arg());
	flag_yield=1;
    break;
  case TOKENIZER_GET_WHEEL:
      accept(TOKENIZER_GET_WHEEL);
      r=get_jogdial_direction();
	  jogdial=0;
	  if (r==0) break;
      r=r==JOGDIAL_LEFT ? -1:1; // left=-1 | right=1
	flag_yield=1;
    break;
  case TOKENIZER_WHEEL_LEFT:
      accept(TOKENIZER_WHEEL_LEFT);
	  if (jogdial==-1) {jogdial=0; r=1; break;} 
	  if (jogdial== 1) {break;} 
      r=get_jogdial_direction();
	  jogdial=r==JOGDIAL_RIGHT; // 
	  if (!r) break;
	  r=r==JOGDIAL_LEFT ? 1:0;
	flag_yield=1;
    break;
  case TOKENIZER_WHEEL_RIGHT:
      accept(TOKENIZER_WHEEL_RIGHT);
	  if (jogdial== 1) {jogdial=0; r=1; break;}
	  if (jogdial==-1) {break;}  
      r=get_jogdial_direction();
	  jogdial=r==JOGDIAL_LEFT; // 
	  if (!r) break;
	  r=r==JOGDIAL_RIGHT;
	flag_yield=1;
    break;
  case TOKENIZER_GET_AFLOCK:
    accept(TOKENIZER_GET_AFLOCK);
	r = shooting_get_prop(camera_info.props.af_lock);
    break;
  case TOKENIZER_GET_AF_FRAME:
	accept(TOKENIZER_GET_AF_FRAME);
    r = camera_info.props.propset==1 ? shooting_get_prop(64) : shooting_get_prop(8);
//	r = shooting_get_prop(camera_info.props.af_frame);
    break;
  case TOKENIZER_GET_FOCUS_MODE:
    accept(TOKENIZER_GET_FOCUS_MODE);
    r = shooting_get_real_focus_mode();
    break;
  case TOKENIZER_GET_FOCUS_STATE:
    accept(TOKENIZER_GET_FOCUS_STATE);
    r = shooting_get_focus_state();
    break;
  case TOKENIZER_GET_FOCUS_OK:
    accept(TOKENIZER_GET_FOCUS_OK);
    r = shooting_get_focus_ok();
    break;
  case TOKENIZER_GET_ZOOM_STEPS:
    accept(TOKENIZER_GET_ZOOM_STEPS);
    r = zoom_points; // ==1  on EOS-M
    break;
  case TOKENIZER_GET_ZOOM:
    accept(TOKENIZER_GET_ZOOM);
    r = shooting_get_zoom();
    break;
  case TOKENIZER_GET_FOCAL_LENGTH:
    accept(TOKENIZER_GET_FOCAL_LENGTH);
    r = get_focal_length(shooting_get_zoom()); //--(lens_get_zoom_point()) wg.  EOS-M
    break;
  case TOKENIZER_GET_EFF_FOCAL_LENGTH:
    accept(TOKENIZER_GET_EFF_FOCAL_LENGTH);
    r = get_effective_focal_length(shooting_get_zoom());
    r = r * conf.zoom_scale / 100; // from gui_osd.c
	break;
//ChangeFocusDistanceToPosition_FW
  case TOKENIZER_GET_FOCUS:
    accept(TOKENIZER_GET_FOCUS);
    subj_dist = shooting_get_subject_distance();
	if (camera_info.dof_values.distance_valid!=0)
    r = subj_dist;
    break;
  case TOKENIZER_GET_SD_OVER_MODES:
    accept(TOKENIZER_GET_SD_OVER_MODES);
    r = sd_over_modes() ; //  0..7
// CAM_SD_OVER_IN_AF           & 1
// CAM_SD_OVER_IN_AFL          & 2 
// CAM_SD_OVER_IN_MF           & 4
    break;
  case TOKENIZER_GET_DRIVE_MODE:
    accept(TOKENIZER_GET_DRIVE_MODE);
    r = shooting_get_drive_mode();
    break;
  case TOKENIZER_SET_MF:
    r = set_mf_statement();
    break;
  case TOKENIZER_GET_MIN_STACK_DIST:
    accept(TOKENIZER_GET_MIN_STACK_DIST);
    r = shooting_get_min_stack_distance();
    break;
  case TOKENIZER_GET_NEAR_LIMIT:
    accept(TOKENIZER_GET_NEAR_LIMIT);
    r = shooting_get_near_limit_of_acceptable_sharpness();
    break;
  case TOKENIZER_GET_FAR_LIMIT:
    accept(TOKENIZER_GET_FAR_LIMIT);
    r = shooting_get_far_limit_of_acceptable_sharpness();
    break;  
   case TOKENIZER_GET_DOF:
    accept(TOKENIZER_GET_DOF);
    r = shooting_get_depth_of_field();
    break;
  case TOKENIZER_GET_HYPERFOCAL_DIST:
    accept(TOKENIZER_GET_HYPERFOCAL_DIST);
    r = shooting_get_hyperfocal_distance();
    break;
  case TOKENIZER_IS_INFINITY:
    accept(TOKENIZER_IS_INFINITY);
	r = (unsigned)subj_dist==INFINITY_DIST;
  //  r = shooting_get_far_limit_of_acceptable_sharpness()==-1;
    break;
  case TOKENIZER_OUT_OF_FOCUS:
	accept(TOKENIZER_OUT_OF_FOCUS);
	if (!eos()) r=expr();
	if (r==-1) {r=subj_dist>0; break;}
	r = subj_dist;
	if (r<=shooting_get_near_limit_of_acceptable_sharpness()) {r=1; break;}
    if (r>=shooting_get_far_limit_of_acceptable_sharpness()) r=-1; else r=0;
	break;
// return true if sensor is enabled (live view on), false if not (playback, rec with display off, display off power save)
  case TOKENIZER_GET_IMAGER_ACTIVE:
    accept(TOKENIZER_GET_IMAGER_ACTIVE);
	r = shooting_get_imager_active() ? 1:0;
	break;
  case TOKENIZER_GET_EV:
    accept(TOKENIZER_GET_EV);
    r = shooting_get_ev_correction1();
    break;
  case TOKENIZER_GET_TV96:
    accept(TOKENIZER_GET_TV96);
    r = shooting_get_tv96();
    break;
  case TOKENIZER_GET_CURRENT_TV96:
    accept(TOKENIZER_GET_CURRENT_TV96);
    r = shooting_get_current_tv96();
    break;
  case TOKENIZER_GET_USER_TV96:
    accept(TOKENIZER_GET_USER_TV96);
    r = shooting_get_user_tv96();
    break;   
  case TOKENIZER_GET_USER_TV_ID:
    accept(TOKENIZER_GET_USER_TV_ID);
    r = shooting_get_user_tv_id();
    break;
  case TOKENIZER_GET_AV96:
    accept(TOKENIZER_GET_AV96);
    r = shooting_get_av96();
    break;  
  case TOKENIZER_GET_CURRENT_AV96:
    accept(TOKENIZER_GET_CURRENT_AV96);
    r = shooting_get_current_av96();
    break;
  case TOKENIZER_GET_USER_AV96:
    accept(TOKENIZER_GET_USER_AV96);
    r = shooting_get_user_av96();
    break;    
  case TOKENIZER_GET_USER_AV_ID:
    accept(TOKENIZER_GET_USER_AV_ID);
    r = shooting_get_user_av_id();
    break;
  case TOKENIZER_GET_MAX_AV96:
    accept(TOKENIZER_GET_MAX_AV96);
    r = shooting_get_max_av96();
    break;
  case TOKENIZER_GET_MIN_AV96:
    accept(TOKENIZER_GET_MIN_AV96);
    r = shooting_get_min_av96();
    break;
  case TOKENIZER_GET_BV96:
    accept(TOKENIZER_GET_BV96);
    r = (int)shooting_get_bv96();
    break;
  case TOKENIZER_GET_SV96:
    accept(TOKENIZER_GET_SV96);
    r = (int)shooting_get_sv96_real();
    break;
  case TOKENIZER_GET_CURRENT_DELTA_SV96:
    accept(TOKENIZER_GET_CURRENT_DELTA_SV96);
    r = shooting_get_current_delta_sv96();
    break;
  case TOKENIZER_GET_CURRENT_BASE_SV96:
    accept(TOKENIZER_GET_CURRENT_BASE_SV96);
    r = shooting_get_current_base_sv96();
    break;
/* from shooting_get_luminance()
  case TOKENIZER_GET_LUMINANCE:
    accept(TOKENIZER_GET_LUMINANCE);
    static const double K = 12.5;               // K is the reflected-light meter calibration constant
    r = (int)shooting_get_bv96();
    r = (int)(100 * 12.5 * pow(2.0,((double)(r-168)/96.0)));
    break;
*/
  case TOKENIZER_GET_ISO_MODE:
    accept(TOKENIZER_GET_ISO_MODE);
    r = shooting_get_iso_mode();
    break;
  case TOKENIZER_GET_ISO_MARKET:
    accept(TOKENIZER_GET_ISO_MARKET);
    r = (int)shooting_get_iso_market();
    break;
  case TOKENIZER_GET_ISO_REAL:
    accept(TOKENIZER_GET_ISO_REAL);
    r = (int)shooting_get_iso_real();
    break;
  case TOKENIZER_GET_EXP_COUNT:
    accept(TOKENIZER_GET_EXP_COUNT);
    r = get_exposure_counter();
    break;
  case TOKENIZER_GET_JPG_COUNT:
    accept(TOKENIZER_GET_JPG_COUNT);
    r = GetJpgCount();
    break;
  case TOKENIZER_GET_RAW_COUNT:
    accept(TOKENIZER_GET_RAW_COUNT);
    r = GetRawCount();
    break;
  case TOKENIZER_GET_RESOLUTION:
    accept(TOKENIZER_GET_RESOLUTION);
    r = shooting_get_resolution();
    break;
  case TOKENIZER_GET_QUALITY:
    accept(TOKENIZER_GET_QUALITY);
    r = shooting_get_prop(camera_info.props.quality);
    break;
  case TOKENIZER_GET_RAW_NR:
    accept(TOKENIZER_GET_RAW_NR);
    r = conf.raw_nr;     //NoiseReduction set 0 if CamError E26
    break;
  case TOKENIZER_GET_FREE_DISK_SPACE:
    accept(TOKENIZER_GET_FREE_DISK_SPACE);
    r = GetFreeCardSpaceKb();
    break;
  case TOKENIZER_GET_PROPSET:
    accept(TOKENIZER_GET_PROPSET);
    r = camera_info.props.propset;
    break;
  case TOKENIZER_GET_PROP:
    accept(TOKENIZER_GET_PROP);
    int var = expr();
    r = shooting_get_prop(var);
    break;
  case TOKENIZER_GET_HISTO_RANGE:
    accept(TOKENIZER_GET_HISTO_RANGE);
    int from = expr();
	accept(-TOKENIZER_COMMA);
    int to = expr();
#ifdef UBASIC_TEST
    (void)from; (void)to;
    r = 0;
#else
    r = (unsigned short)libshothisto->shot_histogram_get_range(from, to);
#endif
    break;
  case TOKENIZER_GET_VBATT:
    accept(TOKENIZER_GET_VBATT);
    r = (unsigned short) stat_get_vbatt();
    break;
  case TOKENIZER_GET_TEMPERATURE:
    accept(TOKENIZER_GET_TEMPERATURE);
    t = expr();
    switch (t)
    {
    case 0: r = get_optical_temp(); 
      break;
    case 1: r = get_ccd_temp(); 
      break;
    case 2: r = get_battery_temp();
      break;
    case 3:
      r = get_battery_temp() + get_ccd_temp() + get_optical_temp(); // Dont panic!
      break;
    default: // do something sane if given a bad index
      r = 0;
    }
    break;
  case TOKENIZER_GET_FLASH_MODE:
    accept(TOKENIZER_GET_FLASH_MODE);
    r = shooting_get_flash_mode();
    break;
  case TOKENIZER_GET_FLASH_READY:
    accept(TOKENIZER_GET_FLASH_READY);
    r = shooting_is_flash();
    break;
  case TOKENIZER_GET_ORIENTATION_SENSOR:
    accept(TOKENIZER_GET_ORIENTATION_SENSOR);
    r = shooting_get_prop(camera_info.props.orientation_sensor);
    break;
  case TOKENIZER_GET_IS_MODE:
    accept(TOKENIZER_GET_IS_MODE);
    r = shooting_get_is_mode();
    break;
  case TOKENIZER_GET_ND_CURRENT_EV96:
    accept(TOKENIZER_GET_ND_CURRENT_EV96);
    r = shooting_get_nd_current_ev96();
    break;
  case TOKENIZER_GET_ND_VALUE_EV96:
    accept(TOKENIZER_GET_ND_VALUE_EV96);
    r = shooting_get_nd_value_ev96();
    break;
  case TOKENIZER_GET_ND_PRESENT:
    accept(TOKENIZER_GET_ND_PRESENT);
    if (camera_info.cam_has_nd_filter == 0)
    {
        r = 0;
    }
    else
    {
        if (camera_info.cam_has_iris_diaphragm == 0)
        {
            r = 1;
        } else {
            r = 2;
        }
    }
    break;
  case TOKENIZER_GET_RAW_SUPPORT:
    accept(TOKENIZER_GET_RAW_SUPPORT);
    r = (is_raw_possible() && !camera_info.state.mode_play);
    break;    
 case TOKENIZER_GET_RAW:
    accept(TOKENIZER_GET_RAW);
#ifdef UBASIC_TEST
    r = 1;
#else
    r = conf.save_raw;
#endif    
    break;
  case TOKENIZER_GET_MODE:
    accept(TOKENIZER_GET_MODE);
    if (camera_info.state.mode_play) r = 1;
    if (!camera_info.state.mode_play && camera_info.state.mode_video) r = 2;
  break;
  case TOKENIZER_GET_METERING_MODE:
    accept(TOKENIZER_GET_METERING_MODE);
	r= shooting_get_prop(camera_info.props.metering_mode);
  break;
  // get CHDK capture mode value, or 0 if in playback or unknown (broken modemap)
  // NOTE: different from get_mode, since this returns the actual value
  case TOKENIZER_GET_CAPTURE_MODE:
    accept(TOKENIZER_GET_CAPTURE_MODE);
    if (camera_info.state.mode_rec) 
      r = camera_info.state.mode_shooting;
//    else r = 0;
    break;
  // check if CHDK capture mode exists in modemap of this camera
  case TOKENIZER_IS_CAPTURE_MODE_VALID:
    accept(TOKENIZER_IS_CAPTURE_MODE_VALID);
    short modenum = expr();
    r = (shooting_mode_chdk2canon(modenum) == -1) ? 0 : 1;
    break;
  case TOKENIZER_GET_VIDEO_BUTTON:
    accept(TOKENIZER_GET_VIDEO_BUTTON);
    r = (camera_info.cam_has_video_button) ? 1 : 0;
    break;
  case TOKENIZER_GET_VIDEO_RECORDING:
    accept(TOKENIZER_GET_VIDEO_RECORDING);
    r = is_video_recording();
    break;
  case TOKENIZER_GET_MOVIE_STATUS:
    accept(TOKENIZER_GET_MOVIE_STATUS);
    r = get_movie_status();
    break;
  case TOKENIZER_GET_CONFIG_VALUE:
    accept(TOKENIZER_GET_CONFIG_VALUE);
    int var1 = expr();
	accept(-TOKENIZER_COMMA);
    int var2 = expr();
    if( conf_getValue(var1, &configVal) == CONF_VALUE) r = configVal.numb; else r = var2;
    break;
  case TOKENIZER_RANDOM:
    accept(TOKENIZER_RANDOM);
    int min = expr();
	accept(-TOKENIZER_COMMA);
    int max = expr();
    // shouldn't srand every time...
    srand((int)shooting_get_bv96()+(unsigned short)stat_get_vbatt()+get_tick_count());
    // wtf
    //action_push_delay(rand()%10);
    flag_yield=1;
    r = min + rand()%(max-min+1);
  break;

  // APEX functions
  case TOKENIZER_ISO_TO_SV96:
    accept(TOKENIZER_ISO_TO_SV96);
    r = shooting_get_sv96_from_iso(expr());
    break;
  case TOKENIZER_SV96_TO_ISO:
    accept(TOKENIZER_SV96_TO_ISO);
    r = shooting_get_iso_from_sv96(expr());
    break;
  case TOKENIZER_ISO_REAL_TO_MARKET:
    accept(TOKENIZER_ISO_REAL_TO_MARKET);
    r = shooting_iso_real_to_market(expr());
    break;
  case TOKENIZER_ISO_MARKET_TO_REAL:
    accept(TOKENIZER_ISO_MARKET_TO_REAL);
    r = shooting_iso_market_to_real(expr());
    break;
  case TOKENIZER_SV96_REAL_TO_MARKET:
    accept(TOKENIZER_SV96_REAL_TO_MARKET);
    r = shooting_sv96_real_to_market(expr());
    break;
  case TOKENIZER_SV96_MARKET_TO_REAL:
    accept(TOKENIZER_SV96_MARKET_TO_REAL);
    r = shooting_sv96_market_to_real(expr());
    break;
  case TOKENIZER_APERTURE_TO_AV96:
    accept(TOKENIZER_APERTURE_TO_AV96);
    r = shooting_get_av96_from_aperture(expr());
    break;
  case TOKENIZER_AV96_TO_APERTURE:
    accept(TOKENIZER_AV96_TO_APERTURE);
    r = shooting_get_aperture_from_av96(expr());
    break;
  case TOKENIZER_USEC_TO_TV96:
    accept(TOKENIZER_USEC_TO_TV96);
    r = shooting_get_tv96_from_shutter_speed((float)expr()/1000000.0);
    break;
  case TOKENIZER_TV96_TO_USEC:
    accept(TOKENIZER_TV96_TO_USEC);
    r = (int)(shooting_get_shutter_speed_from_tv96(expr()) * 1000000.0 + 0.5);
    break;
  case TOKENIZER_SECONDS_TO_TV96:
    accept(TOKENIZER_SECONDS_TO_TV96);
    int n = expr();
	accept(-TOKENIZER_COMMA);
    int d = expr();
    r = shooting_get_tv96_from_shutter_speed((float)n/(float)d);
    break;
  case TOKENIZER_FORCE_ANALOG_AV:
    accept(TOKENIZER_FORCE_ANALOG_AV);
    r=kbd_force_analog_av(expr());
    break;
  case TOKENIZER_FORCE_USB_PRESENT:
    accept(TOKENIZER_FORCE_USB_PRESENT);
    r=force_usb_state(expr());
    break;
  case TOKENIZER_GET_USB_POWER:
    accept(TOKENIZER_GET_USB_POWER);
	short usbpwr = 0;
    if (!eos()) { usbpwr = expr(); }
    r = get_usb_power(usbpwr);
    break;
  case TOKENIZER_GET_ALT_MODE:
    accept(TOKENIZER_GET_ALT_MODE);
    r = (camera_info.state.gui_mode != 0);
    break;    
  case TOKENIZER_GET_DISPLAY_MODE:
    accept(TOKENIZER_GET_DISPLAY_MODE);
    r = shooting_get_display_mode();
    break;
  case TOKENIZER_GET_DRAW_TITLE_LINE:    
    accept(TOKENIZER_GET_DRAW_TITLE_LINE);  
    r = camera_info.state.osd_title_line ;
    break;
  case TOKENIZER_SCRIPT_AUTOSTARTED:
    accept(TOKENIZER_SCRIPT_AUTOSTARTED);
    r = camera_info.state.auto_started;
    break;
  case TOKENIZER_GET_SCRIPT_AUTOSTART:
    accept(TOKENIZER_GET_SCRIPT_AUTOSTART);
#ifdef UBASIC_TEST
    r = 0;
#else
    r = conf.script_startup;
#endif
    break;
  case TOKENIZER_GET_PLATFORM_ID:
    accept(TOKENIZER_GET_PLATFORM_ID);
    r = conf.platformid;
    break;
  case TOKENIZER_GET_DIGIC:
    accept(TOKENIZER_GET_DIGIC);
    r = camera_info.cam_digic;
    break;
  case TOKENIZER_GET_CANON_RAW_SUPPORT:
    accept(TOKENIZER_GET_CANON_RAW_SUPPORT);
    r = camera_info.cam_canon_raw;
    break;
  case TOKENIZER_GET_CANON_IMAGE_FORMAT:
    accept(TOKENIZER_GET_CANON_IMAGE_FORMAT);
    r = shooting_get_canon_image_format();
    break;
  case TOKENIZER_GET_DISK_SIZE:
    accept(TOKENIZER_GET_DISK_SIZE);
    r = GetTotalCardSpaceKb();
    break;
  case TOKENIZER_SWAP_PARTITIONS:
    accept(TOKENIZER_SWAP_PARTITIONS);
    short partNr = expr();
    r = swap_partitions(partNr);
    break;
  case TOKENIZER_GET_DAY_SECONDS:
    {
        accept(TOKENIZER_GET_DAY_SECONDS);
        struct tm *ttm;
        ttm = get_localtime();
        r = ttm->tm_hour * 3600 + ttm->tm_min * 60 + ttm->tm_sec;
    }
    break;
  case TOKENIZER_WAIT_CLICK:
	  r=wait_click_statement(); // r= ms since last click
    break;
  case TOKENIZER_MD_DETECT_TICK:
    accept(TOKENIZER_MD_DETECT_TICK);
    r = camera_info.perf.md_detect_tick;
    flag_yield=1;
    break;
  case TOKENIZER_GET_TIME:
    accept(TOKENIZER_GET_TIME);
    short p=0;
    if (is_token(TOKENIZER_LEFTPAREN)) {p=1; tokenizer_next();}
    if (is_token(TOKENIZER_STRING)) {
        tokenizer_string(string, sizeof(string));
        r=strlen(strpbrk(" SrJTRpYMDhms",string))-1;
        accept(TOKENIZER_STRING); }
    else r=expr();
    if (p) accept(TOKENIZER_RIGHTPAREN);
    r=get_time(r);
    break;
  case TOKENIZER_MSEC:
    accept(TOKENIZER_MSEC);
    r = get_tick_count()-first_tick; // whole line same 
    break;
  case TOKENIZER_LSEC:
    accept(TOKENIZER_LSEC);
    r = get_tick_count()/1000-lsec;
    break;
  case TOKENIZER_SEKUNDE: accept(0);
    r = sec;
    break;
  case TOKENIZER_MINUTE:  accept(0);
    r = sec/60;
    break;
  case TOKENIZER_STUNDE:  accept(0);
    r = sec/3600;
    break;
  case TOKENIZER_PEEK:
    accept(TOKENIZER_PEEK);
	unsigned a=(unsigned)expr(); t=4;
	if (is_comma()) {
		accept(TOKENIZER_COMMA); t=expr()%5; }
	if (a & (t-1)) {accept(TOKENIZER_PEEK); break;}
	if (t==4) r=(unsigned int) *(unsigned int *)(a); else
	if (t==2) r=(unsigned short) *(unsigned short *)(a); else
	if (t==1) r=(unsigned char) *(unsigned char *)(a);
	break;
/*
	long x=0;
	memset(&r, (char)(*(char *)a),1); 
    if (t>1) { x=0;
	memset(&x, (char)(*(char *)(a+1)),1); for big-endian use (a-1) ...
		x<<=8; r+=x;
    if (t>3) {	x=0; 
	memset(&x, (char)(*(char *)(a+2)),1);
		x<<=16; r+=x; x=0;
	memset(&x, (char)(*(char *)(a+3)),1);
		x<<=24; r+=x;	} 
	}
	break;
*/
  case TOKENIZER_VARLIST:
    accept(TOKENIZER_VARLIST);
	r=varlist();
	break;
  case TOKENIZER_STRLEN:
    accept(TOKENIZER_STRLEN);
    r = str_len;
	break;
  case TOKENIZER_LOOP:
    accept(TOKENIZER_LOOP);
	  loop=variables[MAX_VARNUM-1];
      r=loop;	break;
  case TOKENIZER_DEBUG:
    accept(TOKENIZER_DEBUG);
      r=debug;	break;
  case TOKENIZER_BUG:
    accept(TOKENIZER_BUG);
    r = tokenizer_line_number();
    break;
  case TOKENIZER_EXCEPTION: // Romlog on demand?  20..60! secs later
	ended=0; while (ended<1); {ended++;}
	// catch some Errors
  case TOKENIZER_COMMA:
    accept(TOKENIZER_COMMA); // ignore
    r = factor();
	break;
  case TOKENIZER_STRING:
    accept(TOKENIZER_STRING); // ignore
    r = factor();
	break;
  case TOKENIZER_CR:
    //if (debug) 
	accept(TOKENIZER_VARIABLE); return 0; // "Expected 3, got 1"
  case TOKENIZER_ENDOFINPUT:
    accept(TOKENIZER_VARIABLE); return 0;
  case TOKENIZER_ERROR: ended=1;
    accept(TOKENIZER_VARIABLE); return 0;
  case TOKENIZER_TEST: // sandbox for developing
    accept(TOKENIZER_TEST);
	r=expr();
	//r=shooting_get_exif_subject_dist();
	if (is_comma()) {
		accept(TOKENIZER_COMMA); t=expr(); }
	if (ended) {accept(TOKENIZER_TEST); break;}
    break;

  //ARM Begin
      
  default:
    r = 0x7ffffff+1;
    break;
  }
  t=tokenizer_token();
//!  if (t==TOKENIZER_VARIABLE) { ended=1; // "Expected 4, got 3"
//!	  accept(TOKENIZER_PLUS); };  // Breaks some syntax :(
  if (t==TOKENIZER_LEFTPAREN) {// more compat(). to some Lualines.
	 tokenizer_skip(TOKENIZER_RIGHTPAREN);
	 accept(0);} // Caefix
  return r;
}
/*---------------------------------------------------------------------------*/
static long term(void)
{
  long f1, f2, f3=0;
  register short op;
  f1 = factor();
  op = tokenizer_token();
  if (op == TOKENIZER_STRING) {accept(0); // ignore
    op = tokenizer_token();}
  DEBUG_PRINTF("term: token %d\n", op);
  if  ( op == TOKENIZER_CR) return f1;
  while(op == TOKENIZER_ASTR ||
	op == TOKENIZER_SLASH ||
	op == TOKENIZER_EQ ||
	op == TOKENIZER_LT ||
	op == TOKENIZER_GT ||
	op == TOKENIZER_GE ||
	op == TOKENIZER_LE ||
	op == TOKENIZER_NE ||
	op == TOKENIZER_IN ||
	op == TOKENIZER_XOR||
    op == TOKENIZER_OR ||
	op == TOKENIZER_SHL||
	op == TOKENIZER_SHR||
	op == TOKENIZER_MOD )  { tokenizer_next();

	if ( is_token(TOKENIZER_TRUE))
		{f1 = f1!=0;   accept(TOKENIZER_TRUE); return f1;}
	if ( op == TOKENIZER_EQ) {
		if ( is_token(TOKENIZER_SEP))
			{tokenizer_next(); op=TOKENIZER_SEP;} else
		if ( is_token(TOKENIZER_EQ)) tokenizer_next();
	}
	f2 = factor();
    DEBUG_PRINTF("term: %d %d %d\n", f1, op, f2);
    switch(op) {
    case TOKENIZER_LT:     f1 = f1 < f2; break;
    case TOKENIZER_GT:     f1 = f1 > f2; break;
    case TOKENIZER_EQ:     f1 = f1 == f2; break;
    case TOKENIZER_NE:     f1 = f1 != f2; break;
    case TOKENIZER_LE:     f1 = f1 <= f2; break;
    case TOKENIZER_GE:     f1 = f1 >= f2; break;
    case TOKENIZER_ASTR:   f1 *= f2; break;
    case TOKENIZER_OR:     f1 |= f2; break;
    case TOKENIZER_XOR:    f1 ^= f2; break;
    case TOKENIZER_SLASH:  if (!f2) {f1=0xffffffff; break;}
      f1 /= f2; break; // avoid div by zero
    case TOKENIZER_MOD:    if (!f2) {accept(op); break;}
      f1 %= f2; break;
    case TOKENIZER_SHL:    f1 <<=(f2 & 31); break;
    case TOKENIZER_SHR:    f1 >>=(f2 & 31); break;
    case TOKENIZER_IN: { op = f1==f2; 
	  if (is_comma()) do {
         accept(TOKENIZER_COMMA); // ... in 2,4,6,x...
         if (f1 == expr()) op=1; } while (is_comma());
      else { accept(TOKENIZER_TO); f3 = expr();
		 op = (f1>=f2) && (f1<=f3);} // in 2..x
         accept(-TOKENIZER_SEMICOLON);
	  f1 = op; break; }
    case TOKENIZER_SEP:	accept(TOKENIZER_SEMICOLON);
	  f3 = expr(); 	//	 if x =~ 100;4 then ? "x=[96..104] == true"
      f1 = (f1>=f2-f3) && (f1<=f2+f3); break;
    }
    op = tokenizer_token();
  }
  DEBUG_PRINTF("term: %d\n", f1);
  if (op==TOKENIZER_ENDOFINPUT) ended=1;
  return f1;
}
/*---------------------------------------------------------------------------*/
static long expr(void)
{
  long t1, t2;
  register short op;
  
  if (ended) return 0;
  t1 = term();
  op = tokenizer_token();
  if  ( op == TOKENIZER_CR) return t1;
  DEBUG_PRINTF("expr: token %d\n", op);
  while(op == TOKENIZER_PLUS ||
	op == TOKENIZER_MINUS ||
	op == TOKENIZER_AND ||
    op == TOKENIZER_LOR ||
    op == TOKENIZER_NOR ||
	op == TOKENIZER_NAND||
	op == TOKENIZER_LXOR||
	op == TOKENIZER_PRINT)
  
  { tokenizer_next();
    t2 = term();
    DEBUG_PRINTF("expr: %d %d %d\n", t1, op, t2);
   switch(op) {
    case TOKENIZER_PLUS:  t1 += t2; break;
    case TOKENIZER_MINUS: t1 -= t2; break;
    case TOKENIZER_AND:   t1 &= t2; break;
    case TOKENIZER_LOR:   t1 = t1 || t2; break;
    case TOKENIZER_NAND:  t1 = t1 && !t2; break;
    case TOKENIZER_NOR:   t1 = !t1 && !t2; break;// weder noch   
    case TOKENIZER_LXOR:  t1 = !t1 + !t2 ==1; break;
    case TOKENIZER_PRINT: // C style:	I?x:y == wenn I, rechne mit x sonst y
	  if (tokenizer_char() !=  TOKENIZER_PRINT) return t1;
      if (t1) {t1 = t2; accept(TOKENIZER_LABEL); term();} 
	  else { accept(TOKENIZER_LABEL); t1=term();}
      break;
    }
    op = tokenizer_token();
  }
  if (op==TOKENIZER_LAND) {tokenizer_next(); t1= (t1 && relation());}
  DEBUG_PRINTF("expr: %d\n", t1);
  return t1;
}
/*---------------------------------------------------------------------------*/
static long relation(void)
{
  long r1, r2;
  short op;
  
  r1 = expr();
  op = tokenizer_token();
  DEBUG_PRINTF("relation: token %d\n", op);
  while(op == TOKENIZER_LAND || op == TOKENIZER_PRINT) {
    if (op == TOKENIZER_PRINT && tokenizer_char() != TOKENIZER_PRINT) return r1;
    accept(0);
    r2 = expr();
    DEBUG_PRINTF("relation: %d %d %d\n", r1, op, r2);
    switch(op) {
    case TOKENIZER_LAND:
      r1 = r1 && r2;
      break;
    case TOKENIZER_PRINT: // C style:	I?x:y == wenn I, rechne mit x sonst y
      if (r1) {r1 = r2;	accept(TOKENIZER_LABEL); expr();} 
	  else {accept(TOKENIZER_LABEL); r1=expr();}
    }
    op = tokenizer_token();
  }
  return r1;
}
/*---------------------------------------------------------------------------*/
static long varfactor(unsigned short n)
{
  short s=n;
  if (n==MAX_VARNUM) n=tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  if (is_token(TOKENIZER_LEFTPAREN)) { tokenizer_next();
      n=(expr()&63);
      accept(TOKENIZER_RIGHTPAREN);}
  DEBUG_PRINTF("varfactor: obtaining %d from variable %d\n", variables[n&63], n);
  if (s==0x1000) return n;
  if (n==MAX_VARNUM) return 0x7fffffff; // #+
  if (n>MAX_VARNUM) accept(TOKENIZER_VARIABLE);
  return variables[n];
}

/*---------------------------------------------------------------------------*/

// Process a sleep function from the stack
static int action_stack_AS_UBASIC_SLEEP()
{
    if (get_tick_count() >= action_top(2))
    {
        action_pop_func(1);
        return 1;
    }
    return 0;
}
 
/*---------------------------------------------------------------------------*/
static int sleep_delay(int delay)
{
    /* delay of -1 signals indefinite (actually 1 day) delay */
    if (delay == -1)
        delay = 86400000;

    if (delay > 0) {
        return delay + get_tick_count();
    }
    return 0;
}
 
static void sleep_statement(void)
{
    accept(TOKENIZER_SLEEP);
    
	int d=expr();
	int delay = sleep_delay(d);
 
    if (delay > 0)
    {
		if (d==500) {play_sound(5); d=400;} // 
        action_push(delay);
        action_push_func(action_stack_AS_UBASIC_SLEEP);
	//	if ( abs(delay-get_tick_count()) > 10) beep(2); // strange tune.
    } // => get_tick_count updated once per cycle, not per (?).
 
    flag_yield=1;
    DEBUG_PRINTF("End of sleep\n");
    accept_cr();
}
 
static void moment_statement(unsigned d) //:...i....1....;....2....;....3....;....4..
{
	d=d+get_tick_count();
	if (is_token(TOKENIZER_MOMENT)) // '.'
		{accept(0); d=d+50;}
	if (is_token(TOKENIZER_TO)) // '..'
		{accept(0); d=d+200;}
	if (is_token(TOKENIZER_SEKUNDE)) // ''
		{accept(0); d=d+990;}
	action_push(d);
    action_push_func(action_stack_AS_UBASIC_SLEEP);
  flag_yield=1;
}
/*---------------------------------------------------------------------------*/ 
static unsigned stopwatch(int v)  // u=tmode(8) "R"
{
  unsigned u=(get_tick_count()-first_tick);
  sprintf(buf, "--%d : %d:%02d.%02d",v,  u/60000,(u%60000)/1000,u%1000/10);
  script_console_add_error((long)buf);
  return u; 
}
/*---------------------------------------------------------------------------*/
static unsigned get_time(int tmode)
{  int r=42;
   short s,m,h,Y,M,D;
   static struct tm *ttm;
   ttm = get_localtime();
   s=ttm->tm_sec; m=ttm->tm_min; h=ttm->tm_hour; // --~>1000  "SrJTRpYMDhms"
   D=ttm->tm_mday; M=ttm->tm_mon+1; Y=1900+ttm->tm_year;   // "smhDMYpRTJrS"
      if      (tmode==0) r = s;
      else if (tmode==1) r = m;
      else if (tmode==2) r = h;
      else if (tmode==3) r = D;
      else if (tmode==4) r = M;
      else if (tmode==5) r = Y;
	  else if (tmode==6) r = get_tick_count()%1000;
      else {			 r =(get_tick_count()-first_tick); // Runtime
		   if (tmode==7) // buf needs own line
              sprintf(buf, "@ %d:%d : %02d.%02d",sec/3600,sec%3600/60,sec%60,r%1000); else
           if (tmode==8) sprintf(buf, "@ %d/%d/%d # %d:%02d:%02d", Y,M,D,h,m,s); else // Timestamp
           if (tmode==9)  return 1721013+367*Y-7*(Y+(M+9)/12)/4+275*M/9+D; else // Julian day
           if (tmode==10) return r/1000; else
           if (tmode==11) return h * 3600 + m * 60 + s; // GET_DAY_SECONDS

		else {return stopwatch(tmode);}
           // Show timeline, Caefix
           script_console_add_error((long)buf); buf[0]='\0';
		   flag_yield=1;}
    return r;
}

/*---------------------------------------------------------------------------*/
static void writeln()
{
  if (logfile) { 
    FILE *fh = fopen(alog,"ab");
    if (!fh) {
		sprintf(alog, "A/uBerror.Log");
		FILE *fh = fopen(alog,"ab");
		if(!fh) return;}
	fwrite((char *)buf,strlen(buf),1,fh);
	fwrite((char *)"\n",1,1,fh);
	fclose(fh); }
}

static void print_statement(short token)
{
//  static char buf[128];
int i,x=0;
unsigned short u,t; 
short echo=token==TOKENIZER_ECHO;
short pscr=token==TOKENIZER_WRITE;
  buf[0]=0;
  do { accept(0); t=tokenizer_token();} while (t == TOKENIZER_CR);
  if (pscr && t==TOKENIZER_EQ) { // write [=["A/Romlog.log"],]....
    tokenizer_next();
    if (is_token(TOKENIZER_STRING)) {
	  tokenizer_string(alog, sizeof(alog)); accept(0); }
	else sprintf(alog, "A/CHDK/LOGS/Log_%04d.Log", abs(logfile));
	accept(TOKENIZER_COMMA);
	} // 
  if (ended) pscr=0; else	t=tokenizer_token();
  do { string[0]=0; x=0;
    DEBUG_PRINTF("Print loop\n");
	if (ended) break;
    if(t == TOKENIZER_STRING) {
      tokenizer_string(string, sizeof(string));
      sprintf(buf+strlen(buf), "%s", string);
      tokenizer_next();
    } else if(t == TOKENIZER_EQ) {
		tokenizer_inc(); t=tokenizer_token();
		struct tm *ttm = get_localtime();
		switch(t) {
		case TOKENIZER_PLUS:	sprintf(string, "%08d :%02u:%02u:%02u",
		get_tick_count(),ttm->tm_hour, ttm->tm_min, ttm->tm_sec); break;
		case TOKENIZER_SLASH:	get_target_dir_name(string); break;
		case TOKENIZER_ASTR: 	sprintf(string, "%s",scripttitle); break;
		case TOKENIZER_SEMICOLON:sprintf(string, "%s", subtitle); break;
		case TOKENIZER_MOMENT:	sprintf(string, "Img_%04d",get_exposure_counter()); break;
		case TOKENIZER_PRINT:	sprintf(string, "%s-%s-%s-%s-%s",
			 camera_info.platform, camera_info.platformsub, camera_info.chdk_ver,
			 camera_info.build_number, camera_info.build_svnrev);
			break; 
		case TOKENIZER_NOT: 	sprintf(string, "M%d SD%d Dm%d Mm%d F%d AF%dL%d",
			 camera_info.state.mode_shooting, sd_over_modes(), shooting_get_drive_mode(),
			 shooting_get_prop(camera_info.props.metering_mode),
			 shooting_get_real_focus_mode(),  shooting_get_prop(8), // 8=AF_FRAME : 1 = FlexiZone, 2 = Face AiAF / Tracking AF
			 shooting_get_prop(camera_info.props.af_lock));
			break;
		case TOKENIZER_EQ: i=subj_dist;
			 if (i<1) i=shooting_get_exif_subject_dist();
								sprintf(string, "%d <%d> %d", 
		shooting_get_near_limit_of_acceptable_sharpness(), i, shooting_get_far_limit_of_acceptable_sharpness());
			break;
		case TOKENIZER_GT:		sprintf(string, "%d => %d", shooting_get_zoom(),
		get_effective_focal_length(shooting_get_zoom())* conf.zoom_scale / 100);
			break;
		case TOKENIZER_MOD: tokenizer_next();
			x=expr();
			if (x>0xFFFF) {sprintf(string, "0x%08x",x); break;}
			if (x>0xFF)   {sprintf(string, "0x%04x",x); break;}
			sprintf(string, "0x%02x",x); x=-1; break;
		case TOKENIZER_OR: tokenizer_next();
			sprintf(string, "B%d D%d F%d W%d G%d I%d S%d",break_stack_ptr,do_stack_ptr,
			for_stack_ptr,while_stack_ptr,gosub_stack_ptr,if_stack_ptr,select_stack_ptr);
			break;
		// use tok_singlechar to continue...
		default:
			{tokenizer_string(string,123); tokenizer_eol();}
		}
		if ((x==0) && !eos()) tokenizer_next();
		strcat(buf,string);
    } else if(t == TOKENIZER_COMMA) {strcat(buf, " ");
		do {tokenizer_next();} while (is_cr());
    } else if(t == TOKENIZER_LABEL) {tokenizer_next();
		for (u=0; u<=strlen(buf); ++u) {if (buf[u]=='^' && buf[u+1]!='^') buf[u]='"';}
		if (pscr) writeln(); else {
		if (echo) console_add_line(buf); else
		script_console_add_line((long)buf); };
		buf[0]='\0';
		if (eos()) {strcat(buf,":---=----=--+-=----=----"); break;}
    } else if(t == TOKENIZER_SEMICOLON) {
		do {tokenizer_next();
		    if (is_comma()) {strcat(buf, ","); tokenizer_next();}  // ?a;,b == ?a;",";b --> a,b 
		   } while (is_cr());
    } else if(eos()
		|| t == TOKENIZER_MOMENT
		|| t == TOKENIZER_ENDOFINPUT) { break;
    } else {
      sprintf(buf+strlen(buf), "%d", expr());
    }
	if (ended) break;
    t=tokenizer_token();
  } while ( t != TOKENIZER_LET
		&&  t != TOKENIZER_DATA);
  str_len=strlen(buf);
  for (u=0; u<=strlen(buf); ++u)  {if (buf[u]=='^' && buf[u+1]!='^') buf[u]='"';} // "^up^" --> "up"
  
	if (pscr) writeln(); else
	if (echo) console_add_line(buf); else
	if (!is_token(TOKENIZER_STRLEN)) 
		script_console_add_line((long)buf);
  DEBUG_PRINTF("End of print\n");
  
  if ( t == TOKENIZER_MOMENT) accept(0);
  t=tokenizer_token();
  if ( !eos()
	&& t != TOKENIZER_DATA
	&& t != TOKENIZER_LET) accept_cr();
  flag_yield=1; //:...i....1....;....2....;....3....;....4..
}
/*---------------------------------------------------------------------------*/

static void jump_linenum(int linenum)
{
  unsigned short li=ubasic_linenumber();
  if (linenum==li) {return;} // ignore
  if (linenum<li) { linenum=abs(linenum);
	tokenizer_init(program_ptr); }
  while (!is_token(TOKENIZER_NUMBER) || expr() != linenum) {
   /*  Original originell, Speicherjogging?
		//tokenizer_next();
    } //while(tokenizer_token() != TOKENIZER_CR &&
		//	tokenizer_token() != TOKENIZER_ENDOFINPUT);*/
    tokenizer_eol();
	if (is_cr()) {tokenizer_next();}
	else{
	//if (tokenizer_token() == TOKENIZER_ENDOFINPUT)
		if (camera_info.state.state_kbd_script_run == 1) {  
			sprintf(buf, " @%d --> '%d'",li, linenum);
			script_console_add_error((long)buf);
			DEBUG_PRINTF("Label %d not found", linenum);
			ubasic_error = UBASIC_E_UNK_LABEL;
		ended=1; return;}
	}
  DEBUG_PRINTF("jump_linenum: Found line %d\n", tokenizer_num());
  }
}

/*---------------------------------------------------------------------------*/
static void jump_line(unsigned short linenum)
{
  //<< to visited lines only // next, until, wend, return
  if (linenum<ubasic_linenumber()) 
     tokenizer_init(program_ptr);
  while(tokenizer_line_number() < linenum) 
  {tokenizer_eol(); tokenizer_next();}
  /* swallow the CR that would be read next */
  while (is_cr()) tokenizer_next();
///
  if (traconly>9)
  stopwatch(ubasic_linenumber());
}
/*---------------------------------------------------------------------------*/
// TODO: error handling?
int jump_label(char * label)
{
  unsigned short li=ubasic_linenumber();
  char currLabel[MAX_STRINGLEN];
if (go_fwd>=0) {
  if (!go_fwd) tokenizer_init(program_ptr);
  currLabel[0] = 0;
  do  {
	tokenizer_eol(); tokenizer_next();
    if (is_token(TOKENIZER_LABEL)) {
        tokenizer_label(currLabel, sizeof(currLabel));
      if(strcmp(label, currLabel) == 0) {
        DEBUG_PRINTF("jump_linenum: Found line %d\n", tokenizer_line_number());
		tokenizer_eol();
       //= accept(TOKENIZER_CR);
        break;
      }
    }
  } while(!is_eof());
  }
  if (is_eof() || go_fwd<0) {
    if (camera_info.state.state_kbd_script_run == 1) {  
		sprintf(buf, "@%d -> '!%s'",li, label);
		script_console_add_error((long)buf);
      DEBUG_PRINTF("Label %s not found", label);
      ubasic_error = UBASIC_E_UNK_LABEL;
	  go_fwd--; ended=1;}
      return 0;
  } else {
      return 1;
  }
}
/*---------------------------------------------------------------------------*/
static void goto_statement(short token)
{ int ln=0;
  short r=token==TOKENIZER_UBASIC_READER;
  if (r) {pause=10; traconly=100;}
  accept(0);
  while (on_go && !is_cr()) { on_go--;
     if (is_token(TOKENIZER_STRING)) tokenizer_next(); else expr();
     if (is_comma()) tokenizer_next();}
  if (is_cr()) { if (r) return;
		DEBUG_PRINTF("ubasic.c: goto_statement(): no label specified\n");
		ended = 1;
		ubasic_error = UBASIC_E_UNK_LABEL;}
  if (ended) return;
  if (token == TOKENIZER_EXIT) {
      for_stack_ptr = gosub_stack_ptr = while_stack_ptr = do_stack_ptr = 0;
      if_stack_ptr = select_stack_ptr = break_stack_ptr =0;}
  if (is_token(TOKENIZER_STRING)) { ln=0xBA5EBA11;
      tokenizer_string(string, sizeof(string));
      go_fwd = string[0]=='>'; 
	accept(TOKENIZER_STRING);}
  else {ln=expr();}
  if (r && is_comma()) {tokenizer_next(); pause=expr();}
  while (is_comma()) { tokenizer_next(); 
        if (is_token(TOKENIZER_STRING)) tokenizer_next(); else expr();}
  data_statement();
///
  if (r && is_cr()) console_set_layout(0,0,45,12);
  if (ln==0) return; // continue / ignore
  if (ln==0xBA5EBA11) {jump_label(string);} else {jump_linenum(ln);}  // :)
}
/*---------------------------------------------------------------------------*/
static void gosub_statement(void)
{
  int ln=0;
  accept(TOKENIZER_GOSUB);
  while (on_go && !is_cr()) { on_go--;
     if (is_token(TOKENIZER_STRING)) tokenizer_next(); else expr();
     if (is_comma()) tokenizer_next();}
  if (is_cr()) {
		DEBUG_PRINTF("ubasic.c: gosub_statement(): no label specified\n");
		ended = 1;
		ubasic_error = UBASIC_E_UNK_LABEL;}
  if (ended) return;
  short li=tokenizer_line_number();
  if(is_token(TOKENIZER_STRING)) { ln=0x01dfee1d;
    tokenizer_string(string, sizeof(string));
	go_fwd = string[0]=='>'; 
	accept(TOKENIZER_STRING);}
  else {ln=expr();}
  while (is_comma()) { tokenizer_next(); 
        if (is_token(TOKENIZER_STRING)) tokenizer_next(); else expr();}
  data_statement();
  if (ln==0) return; // continue / ignore
  if (gosub_stack_ptr < MAX_GOSUB_STACK_DEPTH) {
/*    tokenizer_line_number_inc();*/
     gosub_stack[gosub_stack_ptr] = li+1;
     gosub_stack_ptr++;
  if (gosub_stack_ptr == MAX_GOSUB_STACK_DEPTH) varlist();
  if (ln==0x01dfee1d) {jump_label(string);} else {jump_linenum(ln);}  // :)
  break_stack[break_stack_ptr] = TOKENIZER_RETURN;
  break_stack_ptr++;
  } else {
      DEBUG_PRINTF("gosub_statement: gosub stack exhausted\n");
      ended = 1;
      ubasic_error = UBASIC_E_GOSUB_STACK_EXHAUSTED;}
}
/*---------------------------------------------------------------------------*/
static void on_statement(void) // on x=0,1,2,3 goto 0,100,20,"3xxx" [,"default"]
{
  int e,v;
  on_go=0;
  accept(TOKENIZER_ON);
  short var = tokenizer_variable_num();
  if (var==-TOKENIZER_LEFTPAREN) {var = varfactor(0x1000);
       v = variables[var];}
  else v = varfactor(var);
  accept(TOKENIZER_EQ);
  do {e=expr();
     if (!is_comma()) break;
     if (e==v) break;
	 on_go++; tokenizer_next();
	 } while (1);
  while (is_comma()) {tokenizer_next(); expr();}
  if (e!=v) on_go++;
}
/*---------------------------------------------------------------------------*/
static void return_statement(void) // breaks loops before, eats endif...
{ short t;
  accept(TOKENIZER_RETURN);
  data_statement();
  if (gosub_stack_ptr > 0) { 
    do { break_stack_ptr--;
	   t=break_stack[break_stack_ptr];
	   if (t == TOKENIZER_NEXT) for_stack_ptr--;
	   if (t == TOKENIZER_UNTIL) do_stack_ptr--;
	   if (t == TOKENIZER_WEND) while_stack_ptr--;
	   if (t == TOKENIZER_ENDIF) if_stack_ptr--;
       if (t == TOKENIZER_END_SELECT) select_stack_ptr--; }
	while (t != TOKENIZER_RETURN);
    gosub_stack_ptr--;
    jump_line(gosub_stack[gosub_stack_ptr]);
  } else {
    DEBUG_PRINTF("return_statement: non-matching return\n");
    ended = 1;
    ubasic_error = UBASIC_E_UNMATCHED_RETURN;
  }
}
/*---------------------------------------------------------------------------*/
static void next_statement(void)
{
  short var, value;
  
  accept(break_stack[break_stack_ptr-1]);  // expected: TOKENIZER_NEXT
  var = tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  if(for_stack_ptr > 0 &&
     var == for_stack[for_stack_ptr - 1].for_variable) {
    value = variables[var] + for_stack[for_stack_ptr - 1].step;
    ubasic_set_variable(var, value);
    
    if(((for_stack[for_stack_ptr - 1].step > 0) && (value <= for_stack[for_stack_ptr - 1].to)) ||
       ((for_stack[for_stack_ptr - 1].step < 0) && (value >= for_stack[for_stack_ptr - 1].to))) {
		data_statement();
		jump_line(for_stack[for_stack_ptr - 1].line_after_for); }
    else {
      for_stack_ptr--;
      break_stack_ptr--;
	  tokenizer_eol();
    //  accept(TOKENIZER_CR);
    }
  } else {
    DEBUG_PRINTF("next_statement: non-matching next (expected %d, found %d)\n", for_stack[for_stack_ptr - 1].for_variable, var);
    ended = 1;
    ubasic_error = UBASIC_E_UNMATCHED_NEXT;
  }

}
/*---------------------------------------------------------------------------*/
static void for_statement(void)
{
  unsigned short for_variable;
  int fo, to, step;
  unsigned short li=tokenizer_line_number();
  accept(TOKENIZER_FOR);
  for_variable = tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  accept(TOKENIZER_EQ);
  fo=expr();
  ubasic_set_variable(for_variable, fo);
  if (is_comma()) {accept(TOKENIZER_COMMA);} else {accept(TOKENIZER_TO);}
  to = expr();                     
  step = 1;
  if (!is_cr() && !is_token(TOKENIZER_DATA)) {
      if (is_comma()) accept(TOKENIZER_COMMA); else accept(TOKENIZER_STEP); ///  :)
	  step = expr();         
  }

  if(for_stack_ptr < MAX_FOR_STACK_DEPTH) {
    for_stack[for_stack_ptr].line_after_for = li+1;
    for_stack[for_stack_ptr].for_variable = for_variable;
    for_stack[for_stack_ptr].to = to;
    for_stack[for_stack_ptr].step = step;
    DEBUG_PRINTF("for_statement: new for, var %d to %d\n",
		 for_stack[for_stack_ptr].for_variable,
		 for_stack[for_stack_ptr].to);
	
    break_stack[break_stack_ptr] = TOKENIZER_NEXT;
    for_stack_ptr++;
    break_stack_ptr++;
	// BUG!?  for i=1 to 0  etc. WILL RUN ONCE in old uBasic! 
  if ((fo>to && step>0) || (fo<to && step<0)) {break_statement(); return;}
  // :...i....1....;....2....;....3....;....4..
  data_statement();
  //accept(TOKENIZER_CR);
  } else {
    DEBUG_PRINTF("for_statement: for stack depth exceeded\n");
    ended = 1;
    ubasic_error = UBASIC_E_FOR_STACK_EXHAUSTED;
  }
}
/*---------------------------------------------------------------------------*/
static void do_statement(void)
{
  if(do_stack_ptr < MAX_DO_STACK_DEPTH) {
     do_stack[do_stack_ptr] = tokenizer_line_number()+1;
     break_stack[break_stack_ptr] = TOKENIZER_UNTIL;
     do_stack_ptr++;
     break_stack_ptr++;
  accept(-TOKENIZER_DO);
  data_statement();
  } else {
    DEBUG_PRINTF("do_statement: do stack depth exceeded\n");
    ended = 1;
    ubasic_error = UBASIC_E_DO_STACK_EXHAUSTED;
  }
}
/*---------------------------------------------------------------------------*/
static void until_statement(void) // do=repeat
{
  int r;
  
  if(do_stack_ptr > 0) {
    accept(break_stack[break_stack_ptr-1]);  // expected: TOKENIZER_UNTIL
    r = relation();
    if(!r) {	data_statement();
      jump_line(do_stack[do_stack_ptr-1]);
    } else {
      do_stack_ptr--;
      break_stack_ptr--;
  	  tokenizer_eol();
    }
  } else {
    DEBUG_PRINTF("until_statement: unmatched until\n");
    ended = 1;
    ubasic_error = UBASIC_E_UNMATCHED_UNTIL;
  }
}
/*---------------------------------------------------------------------------*/
static void while_statement(void)
{
  int r;
  
  accept(TOKENIZER_WHILE);
  if(while_stack_ptr < MAX_WHILE_STACK_DEPTH) {
    if ((while_stack_ptr == 0)
	 ||((while_stack_ptr > 0) && (while_stack[while_stack_ptr-1] != tokenizer_line_number()))){
      while_stack[while_stack_ptr] = tokenizer_line_number();
      break_stack[break_stack_ptr] = TOKENIZER_WEND;
      while_stack_ptr++;
      break_stack_ptr++;
    }
  } else {
    DEBUG_PRINTF("while_statement: while stack depth exceeded\n");
    ended = 1;
    ubasic_error = UBASIC_E_WHILE_STACK_EXHAUSTED;
    return;
  }

  r = relation();
  if(while_stack_ptr > 0) {
    if(!r) {tokenizer_eol(); break_statement();
    } else { data_statement(); }
  } else {
    DEBUG_PRINTF("while_statement: unmatched wend\n");
    ended = 1;
    ubasic_error = UBASIC_E_UNMATCHED_WEND;
  }
}
/*---------------------------------------------------------------------------*/
static void wend_statement(void)
{
  if(while_stack_ptr > 0) {
    accept(break_stack[break_stack_ptr-1]);  // expected: TOKENIZER_WEND
	data_statement();
    jump_line(while_stack[while_stack_ptr-1]);
  } else {
    DEBUG_PRINTF("wend_statement: unmatched wend\n");
    ended = 1;
    ubasic_error = UBASIC_E_UNMATCHED_WEND;
  }
}
/*---------------------------------------------------------------------------*/ 
static void break_statement(void)
{
  register short t,tt,token, i=1;
  short tok = tokenizer_token();
  if (tok ==TOKENIZER_END_SELECT )  {return;}
  short cont= tok==TOKENIZER_CONT;
  if (break_stack_ptr > 0) {
	 while (break_stack[break_stack_ptr-1]==TOKENIZER_ENDIF)
	      { break_stack_ptr--; if_stack_ptr--; }
  }
  if (break_stack_ptr > 0) {
	token=(break_stack[break_stack_ptr-1]);
	 if ( cont || tok==TOKENIZER_BREAK )
		{ accept(0); data_statement(); }
	 if ( token == TOKENIZER_UNTIL) { t=TOKENIZER_DO;}
	 if ( token == TOKENIZER_NEXT ) { t=TOKENIZER_FOR;}
	 if ( token == TOKENIZER_WEND ) { t=TOKENIZER_WHILE;}
	 if ( token == TOKENIZER_END_SELECT ) {t=TOKENIZER_SELECT;}
	 
	 do {	tokenizer_next();
		tt=tokenizer_token();
		if ( tt == t ) { i++; } else 
		if ( tt == token) { i--; } else 
		if ( tt == TOKENIZER_CASE) accept_cr(); else  //	 eat exec {...}
		if ( tt == TOKENIZER_DATA 
		  || tt == TOKENIZER_LET
		  || tt == TOKENIZER_MD_DETECT_MOTION) tokenizer_eol(); else 
		if ( tt == TOKENIZER_ENDOFINPUT) { ended=1; return; }
	} while (i>0);
	
	while (cont) { cont=0;
	  if ( token == TOKENIZER_WEND && while_stack_ptr==0) {token=0; break;} //
	  if ( token == TOKENIZER_NEXT
		|| token == TOKENIZER_WEND 
		|| token == TOKENIZER_UNTIL) { statement();	return;}
	}
	 if ( token == TOKENIZER_END_SELECT ) { return;}
    break_stack_ptr--;
	 if ( token == TOKENIZER_NEXT ) { for_stack_ptr--;}
	 if ( token == TOKENIZER_UNTIL) { do_stack_ptr--;}
	 if ( token == TOKENIZER_WEND ) { while_stack_ptr--;}
	 accept_cr();
  } else {
      DEBUG_PRINTF("break_statement: NO do for while \n");
      ended = 1;
      ubasic_error = UBASIC_E_UNK_LABEL;}
}
/*---------------------------------------------------------------------------*/
/* IF-STATEMENT                                                              */

static void
endif_statement(void)
{
  if(if_stack_ptr > 0) {
    accept(break_stack[break_stack_ptr-1]);  // expected: TOKENIZER_ENDIF
    if_stack_ptr--;
    break_stack_ptr--;
  } else {
    DEBUG_PRINTF("ubasic.c: endif_statement(): endif without if-statement\n");
    ended = 1;
    ubasic_error = UBASIC_E_UNMATCHED_IF;
  }
}
/*---------------------------------------------------------------------------*/
static void else_statement(void)
{
  short r=0, endif_cntr, f_nt, d0;
  
  if(if_stack_ptr > 0) {
    r = if_stack[if_stack_ptr-1];
  }
  else{
    DEBUG_PRINTF("ubasic.c: else_statement(): else without if-statement\n");
	play_sound(0);
    ended = 1;
    accept(TOKENIZER_IF);
    ubasic_error = UBASIC_E_PARSE;
    return;
  }
  accept(TOKENIZER_ELSE);
  DEBUG_PRINTF("else_statement: relation %d\n", r);
  
  d0=(is_token(TOKENIZER_DO));
  if (d0) {if (r) tokenizer_eol();
	  do_statement();}
  if (is_cr()) {
	  accept(TOKENIZER_CR);
	  if(!r) {
      DEBUG_PRINTF("else_statement: result true\n");
      return;
    } else { // eat the rest...
      DEBUG_PRINTF("else_statement: result false\n");
	if (d0) {break_statement(); return;}
      endif_cntr=0;
      while(((!is_token(TOKENIZER_ENDIF ))
           || endif_cntr) && tokenizer_token() != TOKENIZER_ENDOFINPUT){
        f_nt=0;
        if( is_token(TOKENIZER_IF)) { //!tokenizer_skip(TOKENIZER_THEN);
          endif_cntr+=1;
          if (debug)
			{accept(TOKENIZER_IF); relation(); ended=0;} // div by unset var?
        } // short IF
		if( is_token(TOKENIZER_THEN)) { tokenizer_next();
          if (is_token(TOKENIZER_DO))
			 {tokenizer_eol(); do_statement();
			  break_statement(); tokenizer_sep();}
		// then followed by CR -> multi line
          if (is_cr()) { f_nt=1;
          } else { // single line
            endif_cntr--;
            while(!is_token(TOKENIZER_ENDIF) && !is_cr() && !is_eof()){
               tokenizer_next();
            }
            if (is_token(TOKENIZER_ENDIF)){
              DEBUG_PRINTF("ubasic.c: else_statement(): endif in singleline if-statement\n");
              ended = 1;
			  ubasic_error = UBASIC_E_UNMATCHED_IF;
              return;
            }
          }
        }
        if( debug && is_token(TOKENIZER_ELSEIF)) { accept(TOKENIZER_ELSEIF); relation();}
        if( is_token(TOKENIZER_ENDIF))  { endif_cntr--; }
        if (!f_nt) { tokenizer_next(); }
      }
    }
    endif_statement();
  }else{ //statement();

  //  DEBUG_PRINTF("ubasic.c: else_statement(): CR after ELSE expected\n");
  //  ended = 1;
  //  ubasic_error = UBASIC_E_PARSE;
  }
}

/*---------------------------------------------------------------------------*/
static void if_statement(short token)
{
  int r,s=0;
  short t, else_cntr,endif_cntr,f_nt,f_sl;
  
  accept(token); 
	  
  DEBUG_PRINTF("if_statement: get_relation\n");
  r = relation();
  DEBUG_PRINTF("if_statement: relation %d\n", r);
  accept(-TOKENIZER_THEN);
  if (ended) {return;}

  if (is_cr()) { f_nt=1;
    // CR after then -> multiline IF-Statement
    if(if_stack_ptr < MAX_IF_STACK_DEPTH) {
	  if (is_token(TOKENIZER_ELSEIF)) {
		if_stack_ptr--;
        s=if_stack[if_stack_ptr];
      } else {
        break_stack[break_stack_ptr]=TOKENIZER_ENDIF;
        break_stack_ptr++; }
      if (s) r=0; else if_stack[if_stack_ptr] = r;
      if_stack_ptr++;
    } else {
      DEBUG_PRINTF("if_statement: IF-stack depth exceeded\n");
      ended = 1;
      ubasic_error = UBASIC_E_IF_STACK_EXHAUSTED;
      return;
    }
    DEBUG_PRINTF("if_statement: stack_ptr %d\n", if_stack_ptr);
    accept(TOKENIZER_CR);
    if(r) {
      DEBUG_PRINTF("if_statement: result true\n");
      return;

    }else {
      DEBUG_PRINTF("if_statement: result false\n");

      else_cntr=endif_cntr=0; // number of else/endif possible in current nesting
      f_nt=f_sl=0; // f_nt flag for additional next token, f_fs flag single line

	  t=tokenizer_token();
      while( ((t != TOKENIZER_ELSE && t != TOKENIZER_ELSEIF && t != TOKENIZER_ENDIF)
                 || else_cntr || endif_cntr)
			&& t != TOKENIZER_ENDOFINPUT) {
        f_nt=0;
        // nested if
        if( t == TOKENIZER_IF) {  /// if (debug)
			{accept(t); relation(); ended=0;} // div by unset var?
            //!tokenizer_skip(TOKENIZER_THEN);
      //    else_cntr+=1;//;....4..
          endif_cntr+=1;
          f_sl=0;
          DEBUG_PRINTF("IF: line %d, token %d, else %d, end %d\n",
                tokenizer_line_number(),tokenizer_token(),else_cntr,endif_cntr);
          f_nt=1;
          accept( -TOKENIZER_THEN);
          if (tokenizer_token() == TOKENIZER_DO)
			 {tokenizer_eol(); do_statement();
			  break_statement(); tokenizer_sep();}
          DEBUG_PRINTF("THEN: line %d, token %d, else %d, end %d\n",
                tokenizer_line_number(),tokenizer_token(),else_cntr,endif_cntr);
          if (!is_cr()) { f_sl=1; tokenizer_eol();} 
          else {else_cntr+=1;}  //;....4..
          DEBUG_PRINTF("THEN_SL: line %d, token %d, else %d, end %d\n",
                tokenizer_line_number(),tokenizer_token(),else_cntr,endif_cntr);
        } // long IF
        if (is_token(TOKENIZER_ELSE)) {
          else_cntr--;
          DEBUG_PRINTF("ELSE: line %d, token %d, else %d, end %d\n", tokenizer_line_number(),tokenizer_token(),else_cntr,endif_cntr);
          if (else_cntr<0) { 
            DEBUG_PRINTF("ubasic.c: if_statement(): else without if-statement\n");
            ended = 1;
            ubasic_error = UBASIC_E_UNMATCHED_IF;
            return;
          }
        }
        if(!f_sl && (is_token(TOKENIZER_ENDIF))) {tokenizer_eol();
          endif_cntr--;
          if (endif_cntr != else_cntr) else_cntr--;
          DEBUG_PRINTF("ENDIF: line %d, token %d, else %d, end %d\n", tokenizer_line_number(),t,else_cntr,endif_cntr);
        } else {
          if (f_sl && (is_cr()))  {
            f_sl=0;
            endif_cntr--;
            if (endif_cntr != else_cntr) else_cntr--;
            DEBUG_PRINTF("ENDIF_SL: line %d, token %d, else %d, end %d\n", tokenizer_line_number(),t,else_cntr,endif_cntr);
          }else{
            if (is_token(TOKENIZER_ENDIF)) { accept(TOKENIZER_CR)
              DEBUG_PRINTF("ubasic.c: if_statement(): endif in singleline if-statement\n");
              ended = 3;
              ubasic_error = UBASIC_E_PARSE;
              return;}
          }
        }
        if (!f_nt) { tokenizer_next(); }
	  t=tokenizer_token();
      } // while (t !=)...
      if (t == TOKENIZER_ELSE) {else_statement(); return;}
      if (t == TOKENIZER_ELSEIF) {
         if_statement(TOKENIZER_ELSEIF); return;}
    }  // !r
    endif_statement();
  }else {
  // Singleline IF-Statement
  // Now possible: if a=3 then a=33~b=3~c=3 else a=111 ~ b=222~ c=333
	if(r) { //statement(); 
		return;}
//  } else {	// !r
	do  {tokenizer_next();
	   t=tokenizer_token();
		if  (t == TOKENIZER_ELSEIF) { 
             if_statement(TOKENIZER_ELSEIF); return;}
		if  (t == TOKENIZER_ELSE) {accept(TOKENIZER_ELSE); return;}
		if  (t == TOKENIZER_DO)
			{tokenizer_eol(); do_statement();
			break_statement(); tokenizer_sep();} //  select, for & while could follow next line}
	//	if  (t == TOKENIZER_EXEC) {tokenizer_next(); tokenizer_sep();} // eaten by tok.wskip();
	}while (!is_cr() && !is_eof());	// !r
  }
}

/*---------------------------------------------------------------------------*/
/* SELECT-STATEMENT                                                          */
/*---------------------------------------------------------------------------*/
static void
dec_select_stack(void)
{
  if(select_stack_ptr > 0) {
      select_stack_ptr--;
      break_stack_ptr--;
  } else {
    DEBUG_PRINTF("select_statement: SELECT-Stack fail\n");
    ended = 1;
    ubasic_error = UBASIC_E_UNMATCHED_END_SELECT;
  }
}
/*---------------------------------------------------------------------------*/
static void end_select_statement(void)
{
  if(select_stack_ptr > 0) {
	accept(TOKENIZER_END_SELECT);
	tokenizer_eol();
    dec_select_stack();
  } else {
    DEBUG_PRINTF("ubasic.c: end_select_statement(): end_select without select-statement\n");
    ended = 1;
    ubasic_error = UBASIC_E_UNMATCHED_END_SELECT;
  }
}
/*---------------------------------------------------------------------------*/
static void case_statement(void)
{
  int select_value, case_value_1, case_value_2, case_value_eq=1;
  short t, case_run, case_gosub = 0;
  unsigned short cur_ln, gosub_ln = 0;
  
  if(select_stack_ptr > 0) {
    select_value = select_stack[select_stack_ptr - 1].select_value;
    case_run = select_stack[select_stack_ptr - 1].case_run;

    while (!case_run) { 
	t=tokenizer_token();
	if (t == TOKENIZER_CASE_ELSE)
		{case_value_eq=1; accept(TOKENIZER_CASE_ELSE);}
	else {
	  if (t == TOKENIZER_END_SELECT) { break;}
	  accept(TOKENIZER_CASE);
      if (ended) return;
      case_value_1 = expr();
      case_value_eq = (select_value == case_value_1);
      if (case_value_eq) { DEBUG_PRINTF("case_statement: case_value_eq %d, case_value %d\n", case_value_eq, case_value_1); }  

      if(is_token(TOKENIZER_TO)) {
        accept(TOKENIZER_TO);
        case_value_2 = expr();
        if (case_value_1 < case_value_2) {
          case_value_eq = ((select_value >= case_value_1) && (select_value <= case_value_2));
          DEBUG_PRINTF("case_statement: case_value %d to %d\n", case_value_1, case_value_2);
        } else {
          case_value_eq = ((select_value >= case_value_2) && (select_value <= case_value_1));
          DEBUG_PRINTF("case_statement: case_value %d to %d\n", case_value_2, case_value_1);
        }
      } else if (is_comma()) {
        do {
          accept(TOKENIZER_COMMA);
          if (case_value_eq) {
            case_value_2 = expr();
          } else {
            case_value_1 = expr();
            case_value_eq = (select_value == case_value_1);
          }
        } while (is_comma());
        DEBUG_PRINTF("case_statement: case_value_eq %d, case_value_comma %d\n", case_value_eq, case_value_1);
      }
      accept(TOKENIZER_SEMICOLON);
	}  
      if (case_value_eq) {
        //case_gosub = (tokenizer_token() == TOKENIZER_GOSUB);
//GOSUB - save curr linenumber
        cur_ln = tokenizer_line_number();
//GOSUB
		do { t=tokenizer_token();
        switch(t) {
			case TOKENIZER_EXEC: {break;}; // breakstm OK {...here}
			case TOKENIZER_GOSUB:
				{case_gosub = 1; gosub_statement(); break;};
			case TOKENIZER_EXIT:
			case TOKENIZER_GOTO:
				{dec_select_stack(); goto_statement(t); return;};
            default: statement();
            }  // switch
		}
		while (is_token(TOKENIZER_SEP));
//GOSUB  - save new linenumber, reset to curr linenumber
      if (case_gosub) {
        gosub_ln = ubasic_linenumber();
        jump_line(cur_ln);
        DEBUG_PRINTF("case_statement: GOSUB: toLN=%d, nextLN=%d\n", gosub_ln, cur_ln+1);
      }
//GOSUB
        DEBUG_PRINTF("case_statement: case execute\n");
        case_run = 1;
        select_stack[select_stack_ptr - 1].case_run = case_run;
      } else { // !eq
        DEBUG_PRINTF("case_statement: case jump; case_run: %d\n", case_run);
		if  (is_token(TOKENIZER_DO)) {tokenizer_eol();
			do_statement();
			break_statement();} //  select, for & while could follow next line}
		if  (is_token(TOKENIZER_EXEC)) accept_cr(); // eaten by tok.wskip();
        tokenizer_eol();
		tokenizer_next();
      }
//REM
    while (is_cr())	{tokenizer_next();}
//REM
	} // case_run
	  t=tokenizer_token();
      if ((t != TOKENIZER_CASE)
	   && (t != TOKENIZER_EXEC)
	   && (t != TOKENIZER_CASE_ELSE)
       && (t != TOKENIZER_END_SELECT)) {
         DEBUG_PRINTF("ubasic.c: select_statement(): don't found case, case_else or end_select\n");
         ended = 1;
        // ubasic_error = UBASIC_E_PARSE;
		ubasic_error = UBASIC_E_UNMATCHED_END_SELECT;
      } else { 
//GOSUB test for end_select and set to gosub-
        break_statement(); // dont 'break|cont' in script.bas!
        end_select_statement();
        if (case_gosub) {
          gosub_stack[gosub_stack_ptr-1] = tokenizer_line_number()+1; //+1;
          jump_line(gosub_ln);
          DEBUG_PRINTF("end_select_statement: GOSUB: returnLN=%d\n", gosub_stack[gosub_stack_ptr-1]);
        }
//GOSUB        
      }  
  } else { ended = 1;
    DEBUG_PRINTF("case_statement: SELECT-Stack fail\n");
    ubasic_error = UBASIC_E_UNMATCHED_END_SELECT; }
}
/*---------------------------------------------------------------------------*/
static void select_statement(void)
{
  accept(TOKENIZER_SELECT);
  while (on_go) { on_go--;
     expr(); if (is_comma()) tokenizer_next();}
  int select_value = expr(); // I+x
  accept(0);
  if (ended) return;
  
//REM:...,....1....,.
  //  while (tokenizer_token() == TOKENIZER_REM)
	//	{tokenizer_eol(); tokenizer_next();}
//REM
  
  if(select_stack_ptr < MAX_SELECT_STACK_DEPTH) {
    select_stack[select_stack_ptr].select_value = select_value;
    select_stack[select_stack_ptr].case_run = 0;
    DEBUG_PRINTF("select_statement: new select, value %d\n",select_stack[select_stack_ptr].select_value);
	break_stack[break_stack_ptr]=TOKENIZER_END_SELECT;
    select_stack_ptr++;
    break_stack_ptr++;
	
	tokenizer_skip(TOKENIZER_CASE); //:...,....1....,.
    if (!is_token(TOKENIZER_CASE)) {
      DEBUG_PRINTF("ubasic.c: select_statement(): don't found case-statement\n");
      ended = 1;
      //ubasic_error = UBASIC_E_PARSE;
      ubasic_error = UBASIC_E_UNMATCHED_END_SELECT;
    }
    else { }///case_statement(); }
  } else {
    DEBUG_PRINTF("select_statement: SELECT-stack depth exceeded\n");
    ended = 1;
    ubasic_error = UBASIC_E_SELECT_STACK_EXHAUSTED;
  }
}
/* SELECT-STATEMENT END                                                      */
/*---------------------------------------------------------------------------*/

static void let_statement(void)
{
  short var,sep,t;
  long v,x=0;
  long val=0;
  
if (is_token(TOKENIZER_LET)) accept(0);
do {
  var = tokenizer_variable_num();
  if (var==-TOKENIZER_LEFTPAREN) {var = varfactor(0x1000);
       v = variables[var];}
  else v = varfactor(var);
  sep=tokenizer_token();
  switch (sep) {
    case TOKENIZER_EQ: {accept(0);
		 val=expr(); break;}
    case TOKENIZER_PRINT:  accept(0);
      if (v) { val = expr(); accept(TOKENIZER_LABEL); expr();} // C style:	I?x:y == wenn I<>0, I=x sonst I=y
	  else { expr(); accept(TOKENIZER_LABEL); val = expr();}
      break;
    case TOKENIZER_PLUS: { tokenizer_inc();
	 if (is_token(TOKENIZER_EQ)) { tokenizer_next();
		 val=v+expr(); break;}
         accept(TOKENIZER_PLUS);
		 val=v+1; break;}
    case TOKENIZER_MINUS: { tokenizer_inc();
	 if (is_token(TOKENIZER_EQ)) { tokenizer_next();
		 val=v-expr(); break;}
         accept(TOKENIZER_MINUS);
		 val=v-1; break;}
    case TOKENIZER_ASTR: { tokenizer_inc();
	 if (is_token(TOKENIZER_EQ)) { tokenizer_next();
		 val=v*expr(); break;}
         accept(TOKENIZER_ASTR);
		 val=v*v; break;}
    case TOKENIZER_SLASH: {
		 tokenizer_inc(); accept(TOKENIZER_EQ);
		 x=expr();
		 if (x==0) {ended=1; return;}
		 val=v/x; break;}
    case TOKENIZER_MOD: {
		 tokenizer_inc(); accept(TOKENIZER_EQ);
		 x=expr();
		 if (x==0) {ended=1; return;}
		 val=v%x; break;}
    case TOKENIZER_AND: {
		 tokenizer_inc(); accept(TOKENIZER_EQ);
		 val=v&expr(); break;}
    case TOKENIZER_OR: { 
		 tokenizer_inc(); accept(TOKENIZER_EQ);
		 val=v|expr(); break;}
    case TOKENIZER_XOR: { 
		 tokenizer_inc(); accept(TOKENIZER_EQ);
		 val=v^expr(); break;}
    case TOKENIZER_SHL: { 
		 tokenizer_inc(); accept(TOKENIZER_EQ);
		 val=v<<(expr() & 31); break;}
    case TOKENIZER_SHR: { 
		 tokenizer_inc(); accept(TOKENIZER_EQ);
		 val=v>>(expr() & 31); break;}
	default: {
    if (stm_ctr<4 && !is_eof()) { reader_init();} // TOKENIZER_UBASIC_READER
    if (traconly>9) return; 
	
      ended=3;
	  DEBUG_PRINTF("let_statement: Wrong token: '%s'\n",tokenizer_token());
      accept(TOKENIZER_EQ); // "Expected 15, got 3"
      ubasic_error = UBASIC_E_UNK_STATEMENT;
    }
  }
  if (ended) return;
  variables[var]=val;
  DEBUG_PRINTF("let_statement: assign %d to %d\n", variables[var], var);
  t=tokenizer_token();
  sep= t==TOKENIZER_SEP;
  if (sep // print compatibel:
	||t == TOKENIZER_COMMA
	||t == TOKENIZER_SEMICOLON
	||t == TOKENIZER_STRING)	{tokenizer_next();}
  t=tokenizer_token();
  if (t == TOKENIZER_MOMENT
	||t == TOKENIZER_SEKUNDE)	{moment_statement(0);
  t=tokenizer_token();}
  }
  while (t == TOKENIZER_VARIABLE);
  
  sep=sep & (t != TOKENIZER_CR
		  && t != TOKENIZER_ELSE
		  && t != TOKENIZER_ELSEIF);
  if (sep) {tokenizer_sep(); return;}
  if (t == TOKENIZER_CR) return;
  accept_cr();
}
/*---------------------------------------------------------------------------*/
// 10+1 extra vars : #0 - #9 + ##
static void data_statement(void) // to use with goto,gosub,do,for,while,return,break,cont too
{ 
  short var=MAX_VARNUM-13;
  short x=0;
  long val=0;
  while (is_token(TOKENIZER_SEP)) tokenizer_next();
  if (eos()) {tokenizer_eol(); return; }

  short t=tokenizer_token();
  switch (t) {
    case TOKENIZER_PRINT:
    case TOKENIZER_ECHO:
    case TOKENIZER_WRITE:
		 print_statement(t);
	break; }
	if (is_token(TOKENIZER_DATA)) accept(0);
	if (is_token(TOKENIZER_LET)) {let_statement(); return;}
	
	if (is_token(TOKENIZER_EQ))
		{accept(0); var=var+expr()%10;} 
	if (is_token(TOKENIZER_PLUS))
		{accept(0); x=-expr()%10; var=var+x;}
		
	do { var++; x++;
		if (is_comma()) {accept(0);
		if (is_cr()) accept(0);}
		t=tokenizer_token();
		if (t == TOKENIZER_EQ
		  ||t == TOKENIZER_SEMICOLON	// skip 
		  ||t == TOKENIZER_STRING) {accept(0);}
		if (eos()) {tokenizer_eol(); return;} // else 
		{val=expr(); if(x>0) variables[var]=val;}
		if (var>=MAX_VARNUM-3) {tokenizer_eol();}
	}while (is_comma());
  variables[MAX_VARNUM-2]=val; // ##= last set var
  accept_cr();
}
/*---------------------------------------------------------------------------*/
static void shift_statement(void)
{
  short var=MAX_VARNUM-13;
  short i=0, s=1, z=10;
  long val;
  accept(TOKENIZER_SHIFT);
  if (!eos()) s=abs(expr())%10;
  if (is_comma()) {accept(0); z=abs(expr()%10);}
  if (s*z>0)
  for (i=0; i<=z; i++) {
      val=var+i+s; 
      val= (val>=MAX_VARNUM-2) ? 0 : variables[val];
      variables[var+i]=val; 
  } // for 
  accept_cr();
}
/*---------------------------------------------------------------------------*/
static void swap_statement(void)
{
  long v1,v2;
  accept(TOKENIZER_SWAP);
do {
  accept(-TOKENIZER_SEMICOLON);
  short n1=tokenizer_variable_num();
  if (n1==-TOKENIZER_LEFTPAREN) {n1 = varfactor(0x1000);
       v1 = variables[n1];}
  else v1 = varfactor(n1);
  accept(TOKENIZER_COMMA);
  short n2=tokenizer_variable_num();
  if (n2==-TOKENIZER_LEFTPAREN) {n2 = varfactor(0x1000);
       v2 = variables[n2];}
  else v2 = varfactor(n2);
  variables[n1]=v2; 
  variables[n2]=v1; 
  } while (is_token(TOKENIZER_SEMICOLON));
  accept_cr();
}
/*---------------------------------------------------------------------------* /
static void
rem_statement(void)
{
 tokenizer_eol();
 // accept(TOKENIZER_REM);
 // DEBUG_PRINTF("rem_statement\n");
 // accept(TOKENIZER_CR);
}
/ *---------------------------------------------------------------------------*/

static void shoot_statement(void)
{
	unsigned cs;
	short i,w=0;
  accept(TOKENIZER_SHOOT);
  ubasic_as_ret_var_num = -1;
  cs = GetFreeCardSpaceKb();
  if (is_token(TOKENIZER_VARIABLE)) // shoot else failed here
  {
    ubasic_as_ret_var_num = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE); // how to use?
  }
  while (shooting_in_progress()) {moment_statement(42);}  //4..
  action_push_shoot(1);
  //:...i....1....;....2....;....3....;....4..  Caefix
  if ( is_token(TOKENIZER_SEP)) { // rawproof??
	unsigned delay=get_tick_count()+9;
	for (i=0; (i<=conf.save_raw); ++i) {
	  do { 
		delay+=50; w+=50;
		moment_statement(delay);
		if (w>7000) { play_sound(6); break; }
		} while (cs >= GetFreeCardSpaceKb());
	  cs = GetFreeCardSpaceKb();
	  play_sound(4);
	};
  }
  if (cs<42000) {ended=1; loop=off;} // keep some sparespace
  flag_yield=1;
  DEBUG_PRINTF("End of shoot\n");
  accept_cr();
}

/*---------------------------------------------------------------------------*/
short set_mf_statement()
{
   short x;
   accept(TOKENIZER_SET_MF);
   accept(-TOKENIZER_EQ);
   x=(expr() > 0) ? DoMFLock() : UnlockMF();
   accept_cr();
   return x;
}


static void set_focus_statement()
{
    accept(TOKENIZER_SET_FOCUS);
    int sd = expr();// %65536; "some" cams know higher values.
	if (sd==-1) {sd=INFINITY_DIST;}
	if (sd<0) {sd=65536-sd;}
	if (sd==0) {return;} //4....,....
    // if sd override not available now, fail immediately without calling set_focus
    // to avoid unexpected results with SET_LATER
    if(shooting_can_focus()) {
        //// NOTE duplicated in modules/luascript.c and lib/ubasic/ubasic.c
        // in AF lock or MF (canon or set by MF functions), set focus now
        if ( shooting_get_prop(camera_info.props.af_lock) 
          || shooting_get_focus_mode()
          || camera_info.state.mode_video)  // TODO video needs to be investigated, carried over from old code
        {
          shooting_set_focus(sd, SET_NOW);
        }
        else
        {
          // in an AF mode, set later
          shooting_set_focus(sd, SET_LATER);
        }
    }
    accept_cr();
}
/*---------------------------------------------------------------------------*/
static void md_detect_motion_statement()
{
    short columns;
    short rows;
    short pixel_measure_mode;
    int detection_timeout;
    int measure_interval;
    short threshold;
    short draw_grid;
	// var
    short clipping_region_mode=0;
    short clipping_region_row1=0;
    short clipping_region_column1=0;
    short clipping_region_row2=0;
    short clipping_region_column2=0;
    short parameters=0;
    short pixels_step=1;
    short msecs_before_trigger=16;

    accept(TOKENIZER_MD_DETECT_MOTION);

    columns=expr(); tokenizer_next();
    rows=expr(); tokenizer_next();

    pixel_measure_mode=expr(); tokenizer_next();

    detection_timeout=expr(); tokenizer_next();
    measure_interval=expr(); tokenizer_next();

    threshold=expr(); tokenizer_next();

    draw_grid=expr(); tokenizer_next();

    ubasic_as_ret_var_num = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE);
/*
    if (tokenizer_token() != TOKENIZER_CR && tokenizer_token() != TOKENIZER_ELSE) {
        // eat COMA	
        //			tokenizer_next();
    }
*/
	 if (is_comma()) { tokenizer_next();
        clipping_region_mode = expr()%3;
    }
	 if (is_comma()) { tokenizer_next();
        clipping_region_column1 = expr();
    }
	 if (is_comma()) { tokenizer_next();
        clipping_region_row1 = expr();
    }
	 if (is_comma()) { tokenizer_next();
        clipping_region_column2 = expr();
    }
	 if (is_comma()) { tokenizer_next();
        clipping_region_row2 = expr();
    }
	 if (is_comma()) { tokenizer_next();
        parameters = expr();
    }
	 if (is_comma()) { tokenizer_next();
        pixels_step = expr();
    }
	 if (is_comma()) { tokenizer_next();
        msecs_before_trigger = expr();
    }

    accept_cr();

#ifdef UBASIC_TEST
    // TODO: add something here for ubasic_test
#else
    libmotiondetect->md_init_motion_detector(
        columns, rows, pixel_measure_mode, detection_timeout, 
        measure_interval, threshold, draw_grid,
        clipping_region_mode,
        clipping_region_column1, clipping_region_row1,
        clipping_region_column2, clipping_region_row2,
        parameters, pixels_step, msecs_before_trigger);
#endif

    flag_yield=1;
}

/*---------------------------------------------------------------------------*/
static void click_statement(void)
{
  short k;
  accept(TOKENIZER_CLICK);
  k = ubasic_get_key_arg();
  if (k > 0) {
    action_push_click(k);
    flag_yield=1;
  }

  DEBUG_PRINTF("End of click\n");
  accept_cr();
}
/*---------------------------------------------------------------------------*/
static void press_statement(void)
{
  short k;
  accept(TOKENIZER_PRESS);
  k = ubasic_get_key_arg();
  if (k > 0) {
    action_push_press(k);
    flag_yield=1;
  }
  DEBUG_PRINTF("End of press\n");
  accept_cr();
}
/*---------------------------------------------------------------------------*/
static void release_statement(void)
{
  short k;
  accept(TOKENIZER_RELEASE);
  k = ubasic_get_key_arg();
  if (k > 0) {
    action_push_release(k);
    flag_yield=1;
  }
  DEBUG_PRINTF("End of release\n");
  accept_cr();
}
/*---------------------------------------------------------------------------*/
// do
// Wai. 1000
//  K=0 // !! How about that ?  Caefix
//  is_key K "up"   200 ~|goto|data|let..
//  is_key K "set"  500 ~|data|let..
//  is_key K "down" 800 ~|data|let..
// ...   
// until K>0
// goto K*10
//
static void is_key_statement(void)
{ // ++ is_pressed_statement
    short var,val;
	short k=ifkey>-1 && is_token(TOKENIZER_IS_KEY);
	accept(0); flag_yield=1;
    var = tokenizer_variable_num();
  if (var==-TOKENIZER_LEFTPAREN) {var = varfactor(0x1000);
       val = variables[var];}
  else val = varfactor(var);// !!
	if (val) {accept_cr(); tokenizer_eol();	return;} // allready set, so ignore
  //  accept(TOKENIZER_VARIABLE);
  
	if(k) val=script_key_is_clicked(ubasic_get_key_arg()); // 
	else  val=script_key_is_pressed(ubasic_get_key_arg()); // ++
	  
	if (!val) {accept_cr(); tokenizer_eol();} else {
	if (is_token(TOKENIZER_NUMBER)) {val=expr();}
    } 
    ubasic_set_variable(var,val);
    DEBUG_PRINTF("End of is_key\n");
}

/*---------------------------------------------------------------------------*/

static void set_exit_key_statement()
{
    int k;
    accept(TOKENIZER_SET_EXIT_KEY);
    if(is_token(TOKENIZER_STRING)) 
    {
        tokenizer_string(string, sizeof(string));
        k = script_keyid_by_name(string);
        if (k <= 0)
        {
          ubasic_error = UBASIC_E_UNK_KEY;
          ended = 1 ;
        }
        else
        {
            script_set_terminate_key(k,string);
        }
    } 
    else 
    {
       DEBUG_PRINTF("ubasic.c: set_exit_key no key specified\n");
       ended = 1;
       ubasic_error = UBASIC_E_UNK_KEY;
    }
    accept_cr() ;
}

/*---------------------------------------------------------------------------*/
// Wait for a button to be pressed and released (or the timeout to expire)
static int action_stack_AS_UBASIC_WAIT_CLICK()
{
    // Check key pressed or timeout
    if ((get_tick_count() >= action_top(2)) || camera_info.state.kbd_last_clicked)
    {
        // If timed out set key state to "no_key", otherwise key pressed so set last checked time
        if (!camera_info.state.kbd_last_clicked)
            camera_info.state.kbd_last_clicked=0xFFFF;
        else
            camera_info.state.kbd_last_checked_time = camera_info.state.kbd_last_clicked_time;

        action_pop_func(1);
        return 1;
    }
    return 0;
}

static unsigned  wait_click_statement()
{
unsigned ti=get_tick_count();
    accept(TOKENIZER_WAIT_CLICK);
    long delay = -1;
	if (!eos()) {
	if (is_token(TOKENIZER_SEKUNDE)) {accept(0); delay=1000;}
    else { delay = expr();
	  if ( delay == 0) delay = -1;}
    }
    delay = sleep_delay(delay);

    accept_cr();

    // Reset 'clicked' key if it has not changed since last time
    if (camera_info.state.kbd_last_clicked_time <= camera_info.state.kbd_last_checked_time)
    {
        camera_info.state.kbd_last_clicked = 0;
    }

    // Set up for wait or click testing
    action_push(delay);
    action_push_func(action_stack_AS_UBASIC_WAIT_CLICK);

    // Check for short delay or key already pressed by calling action stack routine once now
    if (action_stack_AS_UBASIC_WAIT_CLICK() == 0)
    {
        flag_yield=1;
    }
//  ti=get_tick_count()-ti; // allways 0 !?
  ti=camera_info.state.kbd_last_clicked_time-ti; // nice val ?
  return ti;
}

/*---------------------------------------------------------------------------*/
static void cls_statement(void)
{
  accept(TOKENIZER_CLS);
  console_clear();
  DEBUG_PRINTF("cls_statement\n");
  accept_cr();
}
/*---------------------------------------------------------------------------*/
static void set_console_layout(void)
{
  short x1=0;
  short y1=0;
  short x2=25;
  short y2=5;
  accept(TOKENIZER_SET_CONSOLE_LAYOUT); // 
  x1 = expr();
  if (x1<0) {x1=0; y1=0; x2=45; y2=12;}
  if (is_comma()) {accept(TOKENIZER_COMMA); y1 = expr();}
  if (is_comma()) {accept(TOKENIZER_COMMA); x2 = expr();}
  if (is_comma()) {accept(TOKENIZER_COMMA); y2 = expr();}
  console_set_layout(x1,y1,x2,y2);
  accept_cr();  
}
/*---------------------------------------------------------------------------*/
static void set_console_autoredraw(void)
{
  accept(TOKENIZER_SET_CONSOLE_AUTOREDRAW);
  console_set_autoredraw(expr());
  accept_cr();  
}
/*---------------------------------------------------------------------------*/
static void console_redraw_statement(void)
{
  accept(TOKENIZER_CONSOLE_REDRAW);
  console_redraw();
    accept_cr();
}

static void set_yield_statement()
{
    accept(TOKENIZER_SET_YIELD);
    short val = expr();
	accept(-TOKENIZER_COMMA);
    yield_max_lines = val?val:YIELD_MAX_LINES_DEFAULT;
    val = expr();
    yield_max_ms = val?val:YIELD_MAX_MS_DEFAULT;
    accept_cr();
}

/*---------------------------------------------------------------------------*/

#ifdef INCLUDE_OLD_GET__SYNTAX

static void get_prop_statement()
{
    short var, var1;
    accept(TOKENIZER_GET_PROP);
    var = expr();
	accept(-TOKENIZER_COMMA);
    var1 = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE);
    ubasic_set_variable(var1, shooting_get_prop(var));
    accept_cr();
}

// Call 'func' to get a CHDK short value and store in named parameter
static void get_short_var_statement(short token, short (*func)(void))
{
    accept(token);
    short var = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE);
    ubasic_set_variable(var, func());
    accept_cr();
}

// Call 'func' to get a CHDK int value and store in named parameter
static void get_int_var_statement(short token, int (*func)(void))
{
    accept(token);
    short var = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE);
    ubasic_set_variable(var, func());
    accept_cr();
}

#endif

/*---------------------------------------------------------------------------*/
static void on_off_statement(short token, void (*on)(void), void (*off)(void))
{
  accept(token);
  if (expr() > 0) on(); else off();
  accept_cr();
}
static void switch_function(short token, void (*func)( void))
{
    accept(token); //...4..exit_alt & enter_alt
    func();
    accept_cr();
}
/*
static void on_off_switch(int val, void (*on)(void), void (*off)(void))
{
  if (val > 0) on(); else off();
  //	on_off_switch(0, DoAFLock, UnlockAF);
  //	on_off_switch(1, DoAFLock, UnlockAF);
}
* /
// Call CHDK function 'func' with one int parameter
static int one_int_param_function(int token, void (*func)(int))
{   int r=0;
    accept(token);
    func(r);
    accept_cr();
	return r;
}
*/
static int one_int_param_function(short token, void (*func)(int))
{   int l,r=0;
    accept(token);
	switch(token) { //
		case TOKENIZER_SET_ZOOM: r=expr();
			if (is_comma()) {
				accept(TOKENIZER_COMMA); l=expr();
				shooting_set_zoom_speed(l); }
		break;
		default: r=expr();
	} //
    func(r);
    accept_cr();
	return r;
}

// Call CHDK function 'func' with one short parameter
static void one_short_param_function(short token, void (*func)(short))
{
    accept(token);
    func(expr());
    accept_cr();
}

/*
// Call CHDK function 'func' with one int parameter plus a second value 'p2'
static void one_int_param_plus_const_function(int token, void (*func)(int,short), short p2)
{
    accept(token);
    func(expr(),p2);
    accept_cr();
}

// Call CHDK function 'func' with one short parameter plus a second value 'p2'
static void one_short_param_plus_const_function(int token, void (*func)(short,short), short p2)
{
    accept(token);
    func(expr(),p2);
    accept_cr();
}

/ *---------------------------------------------------------------------------*/

// Set the property 'prop' value from the next expression
static void set_propcase_statement(short token, short prop)
{
    accept(token);
	shooting_set_prop(prop, expr());
    accept_cr();
}

static void set_prop_statement()
{
    int to, to1;
    accept(TOKENIZER_SET_PROP);
    to = expr();
	accept(-TOKENIZER_COMMA);
    to1 = expr();
	shooting_set_prop(to, to1);
    accept_cr();
}

static void set_ev_statement()
{
     int to;
     accept(TOKENIZER_SET_EV);
     to = expr();
         shooting_set_prop(camera_info.props.ev_correction_1, to);
         shooting_set_prop(camera_info.props.ev_correction_2, to);
     accept_cr();
}

/*---------------------------------------------------------------------------*/


static void set_movie_status_statement()
{
    accept(TOKENIZER_SET_MOVIE_STATUS);
    set_movie_status(expr());
    accept_cr();
}


static void set_led_statement()
{
    short to, to1, to2;
    accept(TOKENIZER_SET_LED);
    to = expr();
	accept(-TOKENIZER_COMMA);
    to1 = expr();
	to2 = 200;
	
	if (is_comma()) {accept(TOKENIZER_COMMA); to2 = expr(); }
	camera_set_led(to, to1, to2);
    accept_cr();
}

/*---------------------------------------------------------------------------*/
static void set_capture_mode_canon_statement()
{
    int to;
    accept(TOKENIZER_SET_CAPTURE_MODE_CANON);
    to = expr();
    // if the value as negative, assume it is a mistakenly sign extended PROPCASE_SHOOTING_MODE value
    if( to < 0) 
        to &= 0xFFFF;
    shooting_set_mode_canon(to);
    accept_cr();
}

/*---------------------------------------------------------------------------*/
static void set_config_value_statement()
{
    int id, value;
    tConfigVal configVal = {0};
    
    accept(TOKENIZER_SET_CONFIG_VALUE);
    id = expr();
	accept(-TOKENIZER_COMMA);
    value = expr();
    if( conf_getValue(id, &configVal) == CONF_VALUE ) {
        configVal.numb = value;
        configVal.isNumb = 1;
        conf_setValue(id, configVal);
    }
    accept_cr();
}

static void get_config_value_statement()
{
    int var, var1, var2;
    tConfigVal configVal;
    accept(TOKENIZER_GET_CONFIG_VALUE);
    var = expr();
	accept(-TOKENIZER_COMMA);
    var1 = expr();
	accept(-TOKENIZER_COMMA);
    var2 = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE);
    if( conf_getValue(var, &configVal) == CONF_VALUE ) {
        ubasic_set_variable(var2, configVal.numb);
    } else {
        ubasic_set_variable(var2, var1);
    }
    accept_cr();
}

/*---------------------------------------------------------------------------*/
static void uB_set_av96(int param){
  shooting_set_av96( param, shooting_in_progress()?SET_NOW:SET_LATER );
  return ;
}
static void uB_set_av96_direct(int param){
  shooting_set_av96_direct( param, shooting_in_progress()?SET_NOW:SET_LATER );
  return ;
}
static void uB_set_tv96(int param){
  shooting_set_tv96( param, shooting_in_progress()?SET_NOW:SET_LATER );
  return ;
}
static void uB_set_tv96_direct(int param){
  shooting_set_tv96_direct( param, shooting_in_progress()?SET_NOW:SET_LATER);
  return ;
}
static void uB_set_sv96(int param){
  shooting_set_sv96( param, shooting_in_progress()?SET_NOW:SET_LATER );
  return ;
}
static void uB_set_nd_filter_state(int param) {
   shooting_set_nd_filter_state( param, shooting_in_progress()?SET_NOW:SET_LATER );
  return ;
}
static void uB_set_iso_real(int param) {
  shooting_set_iso_real( param, shooting_in_progress()?SET_NOW:SET_LATER );
  return ;
}
/*---------------------------------------------------------------------------*/

static void md_get_cell_diff_statement()
{
    short var, col, row;
    accept(TOKENIZER_MD_GET_CELL_DIFF);

    col=expr(); tokenizer_next();
    row=expr(); tokenizer_next();

    var = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE);
    accept_cr();

#ifdef UBASIC_TEST
    // TODO: add something here for ubasic_test
#else
    ubasic_set_variable(var, libmotiondetect->md_get_cell_diff(col,row));
#endif
}

static void md_get_cell_val_statement()
{
    short var, col, row;
    accept(TOKENIZER_MD_GET_CELL_VAL);

    col=expr(); tokenizer_next();
    row=expr(); tokenizer_next();

    var = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE);
    accept_cr();

#ifdef UBASIC_TEST
    // TODO: add something here for ubasic_test
#else
    ubasic_set_variable(var, libmotiondetect->md_get_cell_val(col,row));
#endif
}

/*---------------------------------------------------------------------------*/
static int _shot_histogram_set(short enable)
{
#ifdef UBASIC_TEST
    return 0;
#else
    return libshothisto->shot_histogram_set(enable);
#endif
}

/*---------------------------------------------------------------------------*/
static void shutdown_statement(void){
  accept(TOKENIZER_SHUT_DOWN);
  camera_shutdown_in_a_second();
  accept_cr();
}

static void reboot_statement() {
    accept(TOKENIZER_REBOOT);
    if(is_token(TOKENIZER_STRING)) {
        tokenizer_string(string, sizeof(string));
        tokenizer_next();
	    reboot(string);
    } else {
	    reboot(NULL);
    }
}

static void set_autostart_statement()
{
    short to;
    accept(TOKENIZER_SET_SCRIPT_AUTOSTART);
    to = expr();
#ifndef UBASIC_TEST
	if (to >= 0 && to <= 3) conf.script_startup=to;
	conf_save();
#endif
    accept_cr();
}

/*---------------------------------------------------------------------------*/
static void poke_statement(void)
{
    accept(TOKENIZER_POKE);
	unsigned a=(unsigned)expr(); accept(TOKENIZER_COMMA);
	unsigned v=(unsigned)expr(); short s=4;
	if (is_comma()) {
		accept(TOKENIZER_COMMA); s=expr()%5; }
	if (a & (s-1)) {accept(TOKENIZER_POKE); return;}
	if (s==4) *(unsigned int *)(a)  = (unsigned int)(v); else
	if (s==2) *(unsigned short *)(a)= (unsigned short)(v); else
	if (s==1) *(unsigned char *)(a) = (unsigned char)(v);
    accept_cr();
}
/*---------------------------------------------------------------------------*/
static void mem_dump_statement(void)
{
    FILE *fh = fopen("A/mem_dump.Log","wb");
    if(!fh) return;
	unsigned c=(unsigned)expr();
	accept(TOKENIZER_COMMA);
	unsigned s=(unsigned)expr();
	fwrite((char *)c,s,1,fh);
	fclose(fh);
	accept_cr();
}
/*---------------------------------------------------------------------------*/
static void clean_statement(void) // #V..#7 // 42[..62]
{
	accept(TOKENIZER_CLEAN);
	int i;
	short a=0;
	short z=MAX_VARNUM-2; // keep #@ = loop = #(63)
	if (tokenizer_char() == TOKENIZER_VARIABLE) {
		a=tokenizer_variable_num();
		accept(TOKENIZER_VARIABLE);
		accept(TOKENIZER_TO);
		z=tokenizer_variable_num();
		accept(TOKENIZER_VARIABLE);
	} else
	if (!eos()) { a=expr()&63; 
	  if (is_comma() || is_token(TOKENIZER_TO)) tokenizer_next();
	  if (!eos()) z=expr()&63; 
	}
	for (i=a; i<=z; ++i) variables[i]=off;
	accept_cr();
}
/*---------------------------------------------------------------------------*/
static void end_statement(void)
{
  accept(TOKENIZER_END);
  if (!ended & !eos())	end_wait=expr()%10;
	ended = 1;
	loop=off;
}
/*---------------------------------------------------------------------------*/
static void loop_statement()
{
  accept(TOKENIZER_LOOP);
  accept(-TOKENIZER_EQ);
  loop=expr();
  variables[MAX_VARNUM-1] = loop; // = #@ = #(63)
  accept_cr();
}

/*---------------------------------------------------------------------------* /
static void _statement(void)
{
	//:...i....1....;....2....;....3....;....4..
}
/ *---------------------------------------------------------------------------*/
static void test_statement(void)
{
	// accept(TOKENIZER_);
    extern long physw_status; // (use compiler warning as a hint. :) 
 /*
    strcpy(buf, "PS1: %08x",&physw_status[0]);
	int r=(unsigned int) *(unsigned int *)(0x0001a278); 
	sprintf(buf, "PS1: %08x",r);
    script_console_add_error((long)buf);
  //:...i....1....;....2....;....3....;....4..
 */
}
/*---------------------------------------------------------------------------*/
static void list_statement(void)
{
    short l=is_token(TOKENIZER_LIST);
	short s=is_token(TOKENIZER_STRING);
    short d=traconly>9; // (UBASIC_READER);
	if (l) {tokenizer_inc();}
	if (is_token(TOKENIZER_UBASIC_READER)) {stopwatch(-ubasic_linenumber());}
	if (is_cr()) {return;}
	if (l) {tokenizer_sep();}
	tokenizer_string(string,123);
	if (s||traconly) sprintf(buf, "%s", string); else
	sprintf(buf, "@%d$%d: %d '%s'", stm_ctr, ubasic_linenumber(), tokenizer_token(), string);
	script_console_add_error((long)buf); buf[0]='\0';
    if (d && pause==0) msleep(traconly*2);
	if (l) {tokenizer_next();}
	if (is_token(TOKENIZER_STRING) || traconly) tokenizer_eol();
 while (is_token(TOKENIZER_SEP)) tokenizer_next();
 //  //:...i....1....;....2....;....3....;....4..
}

/*---------------------------------------------------------------------------*/ 
void ubasic_reader()
{unsigned short i=1;
while (pause==1 || i==1) { i=0;
  if (!camera_info.state.gui_mode) { pause=-1; return;}  // -->> ubasic_run
  switch (kbd_get_autoclicked_key() | get_jogdial_direction()) {
    case KEY_ZOOM_OUT:
    case JOGDIAL_LEFT:  msleep(100);
         tokenizer_reader_pos(-2); 
         stopwatch(ubasic_linenumber());
         console_redraw(); pause=-1;
        return;
    case KEY_ZOOM_IN:
    case JOGDIAL_RIGHT: msleep(100);
		 if (traconly<100) traconly=100;
         console_redraw(); pause=0;  break;
    case KEY_LEFT:      pause=0; traconly+=50; break;
    case KEY_RIGHT:     pause=2; traconly-=20; 
         tokenizer_reader_pos(-1);   break;
    case KEY_UP:
         pause=12-tokenizer_reader_pos(-24);   break;
    case KEY_DOWN:
         pause=12-tokenizer_reader_pos(24);    break;
    case KEY_SET:
    case KEY_SHOOT_HALF: pause=(pause==-1) ? 12-tokenizer_reader_pos(12) :1;
         if (is_eof()) { pause=8; tokenizer_reader_pos(-8);}  break;
    case KEY_WIFI:
    case KEY_HELP:
    case KEY_FACE:
    case KEY_DISPLAY:    pause=8;
         tokenizer_reader_pos(65432);
         stopwatch(variables[MAX_VARNUM-2]);
         tokenizer_reader_pos(-8);  break;
    case KEY_VIDEO:
    case KEY_PRINT:
    case KEY_RING_FUNC:
         variables[MAX_VARNUM-2]++; // ##++
         stopwatch(variables[MAX_VARNUM-2]);
         play_sound(4); break;
    case KEY_MENU: { if (go_fwd>-2) // continue script ...
		{traconly=TOKENIZER_UBASIC_READER;} else 
        {console_clear(); stopwatch(ubasic_linenumber());}
         pause=-1;
		 return;}     // -->> ubasic_run
    default: {
        if (pause<5) msleep(abs(traconly*2-10*pause));
        }
    } // switch
    if (traconly<10) traconly=10;
    if (traconly>800) traconly=1000;
	if (pause==1 || is_eof()) { pause=-1;
        if (!is_eof()) {console_redraw(); }
         tokenizer_reader_pos(-1);
       // stopwatch(ubasic_linenumber());
    }
    msleep(50);
  } // while
}
/*---------------------------------------------------------------------------*/

static void statement(void)
{

	if (is_token(TOKENIZER_MINUS)) tokenizer_next();
	if (is_token(TOKENIZER_ON))  { on_statement();}
 while (is_token(TOKENIZER_SEP)) { tokenizer_next();
	if (is_cr()) tokenizer_next();}
 while (is_token(TOKENIZER_BEEP)) {play_sound(4); tokenizer_next();}
 while (is_token(TOKENIZER_MOMENT) ||
		is_token(TOKENIZER_SEKUNDE) ||
		is_token(TOKENIZER_TO)) {moment_statement(0);}
 while (is_cr())  {tokenizer_next();}
 
 // Errorology / Reader begin

 if (debug && traconly<9) list_statement(); else
 if (trace) {tokenizer_statement(string,123);
    sprintf(buf, "@%d: '%s'",ubasic_linenumber(), string);
    script_console_add_error((long)buf);
	}

 if (traconly>9) { if (!pause) list_statement();
	 ubasic_reader(); ended=0;
    if (traconly==TOKENIZER_UBASIC_READER) return; // -->> ubasic_run
    if (pause || is_cr()) return;}
 if (ended) return;
 
 
// Strange bug	///here   
/// if ( pause || ended) {return;}
// -->> In file .o/ubasic.elf:
// elf2flt unknown segment 1 for name: 'accept'
// gmake[1]: *** [.o/ubasic.flt] Error 6
// gmake: *** [all-recursive] Error 1

/// if (pause>2) { pause--; return;} // sometimes
// -->> In file .o/ubasic.elf:
// elf2flt unknown symbol: '__gnu_thumb1_case_shi'
// gmake[1]: *** [.o/ubasic.flt] Error 5

// Errorology end   

 
    /*  Fall through. */

  if (is_token(TOKENIZER_NUMBER)) expr();
  ubasic_token token;
  token = tokenizer_token();
  stm_ctr++;
  long l;
  
  switch(token) { 
  case TOKENIZER_VARIABLE:	let_statement();
	if (ended==3) token=TOKENIZER_ERROR;
      break;
  case TOKENIZER_DATA:		data_statement();
      break;
  case TOKENIZER_IF:
  case TOKENIZER_ELSEIF:    if_statement(token);
      break;
  case TOKENIZER_ELSE:		else_statement();
      break;
  case TOKENIZER_ENDIF:		endif_statement();
      break;
  case TOKENIZER_SELECT:	select_statement();
      break;
  case TOKENIZER_CASE:		case_statement();
      break;
  case TOKENIZER_CASE_ELSE: case_statement();
     break;
  case TOKENIZER_END_SELECT:end_select_statement();
     break;
  case TOKENIZER_LET:		let_statement();
      break;
  case TOKENIZER_UBASIC_READER:
      if (traconly>9) {     stopwatch(-stm_ctr); return;}
  case TOKENIZER_EXIT:
  case TOKENIZER_GOTO:		goto_statement(token);
      break;
  case TOKENIZER_GOSUB:		gosub_statement();
      break;
  case TOKENIZER_RETURN:    return_statement();
      break;
  case TOKENIZER_BREAK:		break_statement();
      break;
  case TOKENIZER_CONT:		break_statement();
      break;
  case TOKENIZER_FOR:		for_statement();
      break;
  case TOKENIZER_NEXT:		next_statement();
      break;
  case TOKENIZER_DO:	 		do_statement();
      break;
  case TOKENIZER_UNTIL:		until_statement();
      break;
  case TOKENIZER_WHILE:		while_statement();
      break;
  case TOKENIZER_WEND:		wend_statement();
      break;
  case TOKENIZER_EXEC:		tokenizer_exec();
      if (eos()) accept_cr();
	  break;
  case TOKENIZER_SKIP:
	  accept(TOKENIZER_SKIP);
	  l=tokenizer_token(); accept(0);
	  tokenizer_skip(l);
	  break; // >> Crazy Jump fwd; be careful!
  case TOKENIZER_ECHO:	
  case TOKENIZER_WRITE:	
  case TOKENIZER_STRLEN:
  case TOKENIZER_PRINT:     print_statement(token);
      break;
  case TOKENIZER_CR: return; // -> line_statement 
  case TOKENIZER_SHIFT:		shift_statement();
      break;

  case TOKENIZER_SHOOT:		shoot_statement();
      break;
  case TOKENIZER_CLICK:		click_statement();
      break;
  case TOKENIZER_PRESS:		press_statement();
      break;
  case TOKENIZER_RELEASE:	release_statement();
      break;
  case TOKENIZER_IS_KEY: 	is_key_statement();
      break; // syntax synced :
  case TOKENIZER_IS_PRESSED:is_key_statement();
      break;
  case TOKENIZER_IFKEY:
      accept(TOKENIZER_IFKEY);
      ifkey=expr();
	  accept_cr();
	  break;
  case TOKENIZER_WAIT_CLICK:wait_click_statement();
      break;
  // aelock
  case TOKENIZER_SET_AELOCK:
      on_off_statement(token, DoAELock, UnlockAE);
      break;
  // aflock
  case TOKENIZER_SET_AFLOCK:
      on_off_statement(token, DoAFLock, UnlockAF);
      break;
	  
#ifdef INCLUDE_OLD_GET__SYNTAX
  case TOKENIZER_GET_PROP:
      get_prop_statement();
      break;
  case TOKENIZER_GET_TV96:
      get_short_var_statement(token, shooting_get_tv96);
      break;
  case TOKENIZER_GET_USER_TV96:
      get_short_var_statement(token, shooting_get_user_tv96);
      break;
  case TOKENIZER_GET_USER_TV_ID:
      get_int_var_statement(token, shooting_get_user_tv_id);
      break;
  case TOKENIZER_GET_AV96:
      get_short_var_statement(token, shooting_get_av96);
      break;
  case TOKENIZER_GET_USER_AV96:
      get_short_var_statement(token, shooting_get_user_av96);
      break;
  case TOKENIZER_GET_USER_AV_ID:
      get_int_var_statement(token, shooting_get_user_av_id);
      break;
//#endif
  case TOKENIZER_GET_ZOOM:
      get_int_var_statement(token, shooting_get_zoom);
      break;
  case TOKENIZER_GET_FOCUS:
      get_int_var_statement(token, shooting_get_subject_distance);
      break;
  case TOKENIZER_GET_NEAR_LIMIT:
      get_int_var_statement(token, shooting_get_near_limit_of_acceptable_sharpness);
      break;
  case TOKENIZER_GET_FAR_LIMIT:
      get_int_var_statement(token, shooting_get_far_limit_of_acceptable_sharpness);
      break;
  case TOKENIZER_GET_DOF:
      get_int_var_statement(token, shooting_get_depth_of_field);
      break;
  case TOKENIZER_GET_HYPERFOCAL_DIST:
      get_int_var_statement(token, shooting_get_hyperfocal_distance);
      break;
  case TOKENIZER_GET_ISO_MARKET:
      get_short_var_statement(token, shooting_get_iso_market);
      break;
  case TOKENIZER_GET_ISO_REAL:
      get_short_var_statement(token, shooting_get_iso_real);
      break;
  case TOKENIZER_GET_BV96:
      get_short_var_statement(token, shooting_get_bv96);
      break;
  case TOKENIZER_GET_SV96:
      get_short_var_statement(token, shooting_get_sv96_real);
      break;
  case TOKENIZER_GET_ISO_MODE:
      get_int_var_statement(token, shooting_get_iso_mode);
      break;
  case TOKENIZER_GET_VBATT:
      get_int_var_statement(token, (int (*)(void))stat_get_vbatt);
      break;
  case TOKENIZER_GET_DISK_SIZE:
      get_int_var_statement(token, (int (*)(void))GetTotalCardSpaceKb);
      break;
  case TOKENIZER_GET_FREE_DISK_SPACE:
      get_int_var_statement(token, (int (*)(void))GetFreeCardSpaceKb);
      break;
  case TOKENIZER_GET_JPG_COUNT:
      get_int_var_statement(token, (int (*)(void))GetJpgCount);
      break;
  case TOKENIZER_GET_RAW_COUNT:
      get_int_var_statement(token, (int (*)(void))GetRawCount);
      break;
#endif // old get syntax

  case TOKENIZER_SET_TV96_DIRECT:
      one_int_param_function(token, uB_set_tv96_direct);
      break;
  case TOKENIZER_SET_TV96:
      one_int_param_function(token, uB_set_tv96);
      break;
  case TOKENIZER_SET_USER_TV96:
      one_short_param_function(token, shooting_set_user_tv96);
      break;
  case TOKENIZER_SET_USER_TV_BY_ID:
      one_int_param_function(token, shooting_set_user_tv_by_id);
      break;
  case TOKENIZER_SET_USER_TV_BY_ID_REL:
      one_int_param_function(token, shooting_set_user_tv_by_id_rel);
      break;
  case TOKENIZER_SET_AV96_DIRECT:
      one_int_param_function(token, uB_set_av96_direct );
      break;
  case TOKENIZER_SET_AV96:
      one_int_param_function(token, uB_set_av96 );
      break;
  case TOKENIZER_SET_USER_AV96:
      one_short_param_function(token, shooting_set_user_av96);
      break;
  case TOKENIZER_SET_USER_AV_BY_ID:
      one_int_param_function(token, shooting_set_user_av_by_id);
      break;
  case TOKENIZER_SET_USER_AV_BY_ID_REL:
      one_int_param_function(token, shooting_set_user_av_by_id_rel);
      break;
  case TOKENIZER_SET_SV96:
      one_int_param_function(token, uB_set_sv96 );
      break;
   
  case TOKENIZER_SET_ZOOM:
	if (camera_info.state.gui_mode)
      l=one_int_param_function(token, shooting_set_zoom);
      break;
  case TOKENIZER_SET_ZOOM_REL:
      one_int_param_function(token, shooting_set_zoom_rel);
      break;
  case TOKENIZER_SET_ZOOM_SPEED:
      one_int_param_function(token, shooting_set_zoom_speed);
      break;
  case TOKENIZER_SET_ISO_REAL:
      one_int_param_function(token, uB_set_iso_real );
      break;
  case TOKENIZER_SET_ISO_MODE:
      one_int_param_function(token, shooting_set_iso_mode);
      break;
  case TOKENIZER_SHOT_HISTO_ENABLE:
      one_int_param_function(token, (void (*)(int))_shot_histogram_set);
      break;
  case TOKENIZER_SET_RECORD:
      one_int_param_function(token, shooting_set_playrec_mode);
      break;
  case TOKENIZER_SET_ND_FILTER:
      one_int_param_function(token, uB_set_nd_filter_state );
      break;
  case TOKENIZER_SET_CAPTURE_MODE:
      one_int_param_function(token, (void (*)(int))shooting_set_mode_chdk);
      break;
  case TOKENIZER_SET_CAPTURE_MODE_CANON:
      set_capture_mode_canon_statement();
      break;
  case TOKENIZER_SET_CANON_IMAGE_FORMAT:
      one_int_param_function(token, (void (*)(int)) shooting_set_canon_image_format);
      break;
/*--------------------v-------------------------*/
  case TOKENIZER_EXIT_ALT:
      switch_function(token, exit_alt);
      break;
  case TOKENIZER_ENTER_ALT:
      accept(token);
      enter_alt(1);
      accept_cr();
      break;
/*--------------------^-------------------------*/ 

  case TOKENIZER_SET_FOCUS: set_focus_statement();
      break;
  case TOKENIZER_SET_MF: set_mf_statement();
      break;
  case TOKENIZER_SET_EV: set_ev_statement();
     break;
	   
      // >> mx3 . motion detector
  case   TOKENIZER_MD_DETECT_MOTION:
      md_detect_motion_statement();
      break;
  case  TOKENIZER_MD_GET_CELL_DIFF:
      md_get_cell_diff_statement();
      break;
  case  TOKENIZER_MD_GET_CELL_VAL:
      md_get_cell_val_statement();
      break;
  case TOKENIZER_MD_AF_LED_CONTROL:
      accept(token);
      camera_info.perf.md_af_on_delay = expr();
	  accept(-TOKENIZER_COMMA);
      camera_info.perf.md_af_on_time = expr();
      if ((camera_info.perf.md_af_on_time > 0) && (camera_info.cam_af_led >= 0)) {
          camera_info.perf.md_af_tuning = 1;        // Enable MD testing with AF LED
          // clear previous values
          memset(&camera_info.perf.af_led,0,sizeof(camera_info.perf.af_led));
          camera_info.perf.af_led.min = 0xFFFFFFFF;
      } else {
          camera_info.perf.md_af_tuning = 0;        // Disable MD testing
      }
      accept_cr();
      break;
      // << mx3 . motion_detector

/*
  case TOKENIZER_SET_ISO_MARKET:
      one_int_param_function(token, shooting_set_iso_market);
      break;
  case TOKENIZER_SET_ISO_DL_F_B:
      one_int_param_function(token, shooting_set_iso_real_delta_from_base);
      break;
*/  
  //ARM End
  
  case TOKENIZER_SET_PROP: set_prop_statement();
      break;
  case TOKENIZER_SET_LED: set_led_statement();
      break;

  case TOKENIZER_SET_MOVIE_STATUS:
      set_movie_status_statement();
      break;
  case TOKENIZER_SET_RESOLUTION:
      set_propcase_statement(token, camera_info.props.resolution);
      break;
  case TOKENIZER_SET_QUALITY:
      set_propcase_statement(token, camera_info.props.quality);
      break;
  case TOKENIZER_SET_METERING_MODE:
 	  set_propcase_statement(token, camera_info.props.metering_mode);
      break;
  case TOKENIZER_SET_AF_FRAME: //?? 
	  accept(TOKENIZER_SET_AF_FRAME);
	  l=expr(); // %3
      if (camera_info.props.propset==1) shooting_set_prop(64,l);
	  else shooting_set_prop(8,l);
//      shooting_set_prop(camera_info.props.af_frame, l); // 1 = FlexiZone, 2 = Face AiAF / Tracking AF); NO change!
      accept_cr();
    break;


  case TOKENIZER_SET_RAW: accept(token);
      conf.save_raw = expr();
      accept_cr();
      break;
  case TOKENIZER_SET_RAW_NR: accept(token);
      conf.raw_nr = expr();
      accept_cr();
      break;
  case TOKENIZER_SET_SCRIPT_AUTOSTART:
      set_autostart_statement();
      break;
      
  case TOKENIZER_GET_CONFIG_VALUE:
    get_config_value_statement();
    break;

  case TOKENIZER_SET_CONFIG_VALUE:
    set_config_value_statement();
    break;
  case TOKENIZER_SET_YIELD:
    set_yield_statement();
    break;

  case TOKENIZER_USB_SYNC_WAIT:
    accept(TOKENIZER_USB_SYNC_WAIT);
#ifdef UBASIC_TEST
    // TODO: add something here for ubasic_test
#else
    if (expr()) usb_sync_wait_flag = 1;
    else        usb_sync_wait_flag = 0;
#endif
    accept_cr();
    break;

  case TOKENIZER_SET_REMOTE_TIMING:
    accept(TOKENIZER_SET_REMOTE_TIMING);
#ifdef UBASIC_TEST
    // TODO: add something here for ubasic_test
#else
    int hpenable= expr();
    if ( hpenable > 0) start_usb_HPtimer(hpenable);
    else stop_usb_HPtimer();
#endif
    accept_cr();
    break;

  case TOKENIZER_SET_EXIT_KEY:
      set_exit_key_statement();
      break;

  case TOKENIZER_WHEEL_LEFT:
      accept(token);
      JogDial_CCW();
      accept_cr();
      break;
  case TOKENIZER_WHEEL_RIGHT:
      accept(token);
      JogDial_CW();
      accept_cr();
      break;

  case TOKENIZER_CLS: cls_statement();
      break;
  case TOKENIZER_SET_CONSOLE_LAYOUT:
      set_console_layout();
      break;
  case TOKENIZER_SET_CONSOLE_AUTOREDRAW:
      set_console_autoredraw();
      break;
  case TOKENIZER_CONSOLE_REDRAW:
      console_redraw_statement();
      break;
  case TOKENIZER_SET_DRAW_TITLE_LINE:
      accept(token);
      camera_info.state.osd_title_line=expr()?1:0;
      accept_cr();
      break;
  case TOKENIZER_SET_BACKLIGHT:
      on_off_statement(token, TurnOnBackLight, TurnOffBackLight);
      break;
  case TOKENIZER_SET_LCDDISPLAY:
      on_off_statement(token, TurnOnDisplay, TurnOffDisplay);
      break;
  case TOKENIZER_PRINT_SCREEN:
      l=one_int_param_function(token, script_print_screen_statement);
	  if (l) {logfile=l;
	  sprintf(alog, "A/CHDK/LOGS/Log_%04d.Log", abs(logfile));}
	  flag_yield=1;
      break;
  case TOKENIZER_PLAY_SOUND:
      one_int_param_function(token, (void (*)(int))play_sound);
      break;
  case TOKENIZER_BEEP:
	  {play_sound(4); accept(TOKENIZER_BEEP);}
	  break;
  case TOKENIZER_SLEEP: sleep_statement();
      break;
  case TOKENIZER_SWAP: swap_statement();
      break;
  case TOKENIZER_LIST:
//	  accept(TOKENIZER_LIST);
      list_statement();
	  break;
  case TOKENIZER_VARLIST:
		accept(TOKENIZER_VARLIST);
		if (is_token(TOKENIZER_STRING)) list_statement();
      varlist();
      break;
  case TOKENIZER_POKE: poke_statement();
      break;
  case TOKENIZER_MEM_DUMP:
		accept(TOKENIZER_MEM_DUMP);
		mem_dump_statement();
      break;
	  
  case TOKENIZER_LOOP: loop_statement();
	  break;
  case TOKENIZER_TRONLY:
	tokenizer_next(); trace=on; traconly=on; break;
  case TOKENIZER_TRON:
	tokenizer_next(); trace=on; traconly=off;break;
  case TOKENIZER_TROFF: trace=off; traconly=off;
	tokenizer_next();
      break;
  case TOKENIZER_RESET: 
	accept(TOKENIZER_RESET);  data_statement();
	  ended=-1;
      break;
  case TOKENIZER_REBOOT:
      reboot_statement();
      break;
  case TOKENIZER_SHUT_DOWN:
      shutdown_statement();
  case TOKENIZER_ENDOFINPUT: ended=1; traconly=off;
		break;
  case TOKENIZER_EXCEPTION: // Romlog on demand? 20-60! secs later
	ended=0; while (ended<1); {ended++;}
  case TOKENIZER_DEBUG:
    accept(TOKENIZER_DEBUG);
    accept(-TOKENIZER_EQ);
      if (eos()) debug=on; else debug=expr();
	  accept_cr();
	  break;
  case TOKENIZER_TEST:
    accept(TOKENIZER_TEST);
	test_statement();
      break;
  case TOKENIZER_CLEAN:
    clean_statement();
	  break;
  case TOKENIZER_BEGIN:
    tokenizer_begin();
	reset();
      break;
  case TOKENIZER_BUG:
    accept(TOKENIZER_CR); ended=3;
      break;
  case TOKENIZER_END: end_statement();
      break;
//case TOKENIZER_REM: tokenizer_eol(); break;
	   
  default: {
    if (stm_ctr<4 && !is_eof()) {reader_init();}
    if (traconly>9) return; ///{ stopwatch(ubasic_linenumber());}

      DEBUG_PRINTF("ubasic.c: statement(): not implemented %d\n", token);
      ended = 3;  // Letters eaten in let_statement
      print_statement(TOKENIZER_PRINT);
      ubasic_error = UBASIC_E_UNK_STATEMENT;
    }
  }// break>>
  l=tokenizer_token();
  if (l==TOKENIZER_ELSE || l==TOKENIZER_ELSEIF) accept_cr(); /// eat singleline if ...[exec {...}]
  
  sec=(get_tick_count()-first_tick)/1000;
}

/*---------------------------------------------------------------------------*/
static void line_statement(void)
{
  /* line numbers have been removed */
  DEBUG_PRINTF("----------- Line number %d ---------\n", tokenizer_line_number());
  /*    current_linenum = tokenizer_num();*/
#if 0
  if (is_token(TOKENIZER_LABEL)) {
#ifdef DEBUG
  if (traconly<9)      tokenizer_label(string, sizeof(string));
      DEBUG_PRINTF("line_statement: label: %s\n", string );
#endif
      //accept(TOKENIZER_LABEL);

  if (traconly>9) { tokenizer_eol(); accept(0);
      return; }
  }
#endif 
  /* reyalp - eat up to 100 labels or rems at a time so they don't cost 10ms each */
  /* Caefix - 10ms are a hollyday for a Canon, lets use that! */
  short count = 100;

  if (traconly>9) { // ubasic_reader
     if (!camera_info.state.gui_mode) { 
         pause=-2; return;}
     if (pause==-2 && camera_info.state.gui_mode) { enter_alt(1);
         kbd_get_autoclicked_key();
         console_redraw(); flag_yield=1; pause=-1; return;}
  }
  else do {
    short r =  tokenizer_token();
    if ( r == TOKENIZER_LABEL ) {
      /* hit limit and we are on a label, return */
      if( count == 1 ) return;
#ifdef DEBUG
      tokenizer_label(string, sizeof(string));
      DEBUG_PRINTF("line_statement: label: %s\n", string );
#endif
	  tokenizer_eol();
    }
    else if ( r == TOKENIZER_VARIABLE || // # & Lets nice here!
			  r == TOKENIZER_LET) { let_statement(); }
	// lets eat what we dont need on 1. Position.
    else if ( r == TOKENIZER_MINUS) { tokenizer_next();
		 if ( is_token(TOKENIZER_MINUS)) tokenizer_eol();} // skip Lua-commentline
    else if ( r == TOKENIZER_SLASH ) {tokenizer_eol();} // skip C-commentline
    else if ( r == TOKENIZER_STRING ) {list_statement(); tokenizer_eol();} // show "commentline"
	else if ( r == TOKENIZER_SEP ) {tokenizer_next(); break;} // no Yield
    else if ( r == TOKENIZER_ASTR ) { accept(0); int x=0;
		switch(tokenizer_token()) {  // '*' 
			case TOKENIZER_SEP: tokenizer_next(); break;
			case TOKENIZER_MOMENT: ended=1; return;
			case TOKENIZER_LABEL: accept(0); get_time(7); break;
			case TOKENIZER_PLUS: accept(0); get_time(42); break;
			case TOKENIZER_LNOT: accept(0); traconly=off; break;
			case TOKENIZER_GT: { accept(0);
			   x=tokenizer_token(); accept(0);
				 tokenizer_skip(x); break;} // Crazy Jump; be careful!
			case TOKENIZER_LEFTPAREN: { // *(
			  do { tokenizer_eol(); accept(0); // ignore multilineblock *( <# # #> \n*)
				 if (is_token(TOKENIZER_ASTR)) {accept(0);
				 x=is_token(TOKENIZER_RIGHTPAREN);}
				} while (!(x|ended));
				if (x) tokenizer_sep(); } // *)
				
			default: {tokenizer_eol();}
		}
	} // if '*' 
	else if ( r == TOKENIZER_NUMBER) {expr();} //ignore: numeric label for goto|gosub 10*x
	else if ( r ==TOKENIZER_LEFTPAREN) {
		tokenizer_skip(TOKENIZER_RIGHTPAREN);
		accept(TOKENIZER_RIGHTPAREN);}
	else if ( r ==TOKENIZER_RIGHTPAREN) {tokenizer_next();}
	else if ( r == TOKENIZER_ENDOFINPUT) {return;} // -> ubasic_run
	else {	
		if (is_cr()) { accept(0); continue; }
		break; }
  } while(--count);
  
 do { statement();	if (flag_yield|ended) break;
	} while (is_token(TOKENIZER_SEP));
	
 if (debug) {
	tokenizer_statement(string,120);
    sprintf(buf, "@%d: '%s'",ubasic_linenumber(), string);
    script_console_add_error((long)buf);
	}
  sec=(get_tick_count()-first_tick)/1000;

  // ubasic_reader
  if (traconly>9) {
	 if (pause>0) { pause--;
	    if (!is_eof()) {list_statement(); ended=0;}
     }
  } /**/
  return;
}
/*---------------------------------------------------------------------------*/
int ubasic_run(void)
{
	unsigned short li,i,lines = 0;
    flag_yield = 0;
    start_tick = get_tick_count();
	li=ubasic_linenumber(); // catch jumpline for error
 //   subj_dist=shooting_get_subject_distance();
  
    do
    {     
        if ( ended )
        {
		//  Generate [error] message
            char buf[123];
            const char *msg;
            DEBUG_PRINTF("%ds uBASIC program finished\n",sec);
			i = (start_sd-GetFreeCardSpaceKb())/1024;
			sprintf(buf, "-%dMb %d @%d Log %d",i, GetFreeCardSpaceKb()/1024, stm_ctr,logfile);
			script_console_add_error((long)buf);
			get_time(8);
			if ( ended==-1) {reset(); break;}
            if (ubasic_error)
            {
                if (ubasic_error >= UBASIC_E_ENDMARK)
                {
                    msg = ubasic_errstrings[UBASIC_E_UNKNOWN_ERROR];
                }
                else
                {
                    msg = ubasic_errstrings[ubasic_error];
                }
                sprintf(buf, "%d:%d uBASIC:%s", ubasic_linenumber(), tokenizer_token(), msg);
                // Show error message
                script_console_add_error((long)buf);
				
                // Show error line, Caefix
				tokenizer_statement(string,50);
				sprintf(buf, "'%s'", string);
                script_console_add_error((long)string);
				if (ended==3) li= ubasic_linenumber();
				jump_line(li);
				list_statement();
			script_console_add_error(LANG_CONSOLE_TEXT_TERMINATED); buf[0]='\0';
				for (i=0; i<=5; i++) moment_statement(990);
				if (loop) {reset(); return SCRIPT_RUN_RUNNING;}
			console_set_layout(0,0,25,5);
            return SCRIPT_RUN_ERROR;
            }
            // Show 'Finished' message//
			sprintf(buf, "%d:%d uBASIC: %04d",
              ubasic_linenumber(), tokenizer_token(), get_exposure_counter());
			script_console_add_error((long)buf);
            script_console_add_error(LANG_CONSOLE_TEXT_FINISHED);
			if (loop) {reset();
			for (i=0; i<=end_wait; i++) {moment_statement(990);}
			return SCRIPT_RUN_RUNNING;}
			console_set_layout(0,0,25,5);
            return SCRIPT_RUN_ENDED;
        }

		while (is_cr())  {tokenizer_next();}
        line_statement();

		if (traconly==TOKENIZER_UBASIC_READER) { go_fwd=1;
            pause=ubasic_linenumber();
            if (!jump_label("BEGIN")) { tokenizer_init(program_ptr);
            if (!jump_label("BEGIN")) { jump_line(pause-2);
               stopwatch(-ubasic_linenumber());}
            }
		    traconly=pause=ended=0; break;
        }

        if ( flag_yield && !ended) break; // eat eol here.
		if (is_cr() || ended) continue;

        // Return control to CHDK only if external processing required  
		// && to get updated values like get_tick_count ? //Caefix
        if ( flag_yield )
            return SCRIPT_RUN_RUNNING;

        lines++;
    } while (lines < yield_max_lines && get_tick_count() - start_tick < yield_max_ms);
	lsec=(get_tick_count()-first_tick)/1000-lsec;
    return SCRIPT_RUN_RUNNING;
}
/*---------------------------------------------------------------------------* /
static long ubasic_get_variable(short varnum)
{
  if (varnum >= 0 && varnum < MAX_VARNUM) 
    return variables[varnum];

  return 0;
}
/ *----- 4051 -----------------------------------------------------------------*/
// read a key name and return key id, 
// set error and return 0 if key invalid
static short ubasic_get_key_arg()
{
  short k,p=0;
  if (is_token(TOKENIZER_LEFTPAREN)) {p=1; tokenizer_next();}
  if (is_token(TOKENIZER_STRING)) {
     tokenizer_string(string, sizeof(string));
     accept(TOKENIZER_STRING);
     if (strncmp(string,".", 2)==0) {k=mem_key;}
     else {k=script_keyid_by_name(string); mem_key=k;}
  } else k=expr();  // NO errortest!
  if (p) accept(TOKENIZER_RIGHTPAREN);
  if (k <= 0) ubasic_error = UBASIC_E_UNK_KEY;
  return k;
}
/*---------------------------------------------------------------------------*/

// Save Action Stack 'return' value in selected variable
void ubasic_set_as_ret(int md_ret);
inline void ubasic_set_as_ret(int md_ret)
{
 //   if (ubasic_as_ret_var_num >= 0)  // Letters only
        ubasic_set_variable(ubasic_as_ret_var_num, md_ret);
}
/*---------------------------------------------------------------------------*/
void ubasic_end(void)
{
	tokenizer_end();
//	msleep(1000);
//	moment_statement(500);
}
// https://app.assembla.com/spaces/chdk/subversion/commits/5792
/* following .bas script starts with same variable values.
/*----- 4086 -------------- 5910 ---------------------------------------------*/